/*
 * $Id: VMSymbol.c 210 2008-04-16 12:54:12Z michael.haupt $
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */
 
#include "VMSymbol.h"

#include <memory/gc.h>

#include <string.h>

//
//  Class Methods (Starting with VMSymbol_) 
//
//

/**
 * Create a new VMSymbol
 */
pVMSymbol VMSymbol_new(void) {
    pVMSymbol result = (pVMSymbol)gc_allocate_object(sizeof(VMSymbol));
    if(result) {
        result->_vtable = VMSymbol_vtable();
        gc_start_uninterruptable_allocation();
        INIT(result, NULL);
        gc_end_uninterruptable_allocation();
    }
    return result;
}

/**
 * Create a new VMSymbol with an initial string
 */
pVMSymbol VMSymbol_new_from_string(pString string) {
    pVMSymbol result = (pVMSymbol)gc_allocate_object(sizeof(VMSymbol));
    if(result) {
        result->_vtable = VMSymbol_vtable();
        gc_start_uninterruptable_allocation();
        INIT(result, string);
        gc_end_uninterruptable_allocation();
    }
    return result;
}

/**
 * Create a new VMSymbol with an initial C-string
 */
pVMSymbol VMSymbol_new_from_chars(const char* restrict string) {
    pVMSymbol result = (pVMSymbol)gc_allocate_object(sizeof(VMSymbol));
    if(result) {
        result->_vtable = VMSymbol_vtable();
        gc_start_uninterruptable_allocation();
        INIT(result, String_new(string));
        gc_end_uninterruptable_allocation();
    }
    return result;
}


/**
 * Initialize a VMSymbol
 */
void _VMSymbol_init(void* _self, ...) {
    pVMSymbol self = (pVMSymbol)_self;    
    SUPER(VMObject, self, init, 0);
    
    va_list args;
    va_start(args, _self);
    pString embed = String_new_from(va_arg(args, pString));
    va_end(args);
    if(embed)
        SEND(self, set_string, embed);
    else
        self->embedded_string = NULL;
    self->plain_string = NULL;
}


void _VMSymbol_free(void* _self) {
    pVMSymbol self = (pVMSymbol)_self;    
    
    if(self->embedded_string)
        SEND(self->embedded_string, free);
    SUPER(VMObject, self, free);
}


//
//  Instance Methods (Starting with _VMSymbol_) 
//
//
pString _VMSymbol_get_string(void* _self) {
    pVMSymbol self = (pVMSymbol)_self;
    return self->embedded_string;
}


// embedded string will be duplicated
void _VMSymbol_set_string(void* _self, pString embedded) {
    pVMSymbol self = (pVMSymbol)_self;
    self->embedded_string = embedded;
    self->hash = embedded->hash;        
}


pString _VMSymbol_get_plain_string(void* _self) {
    pVMSymbol self = (pVMSymbol)_self;
    if(self->plain_string)
        return self->plain_string;
    
    self->plain_string = String_new("");
    
    for(size_t i=0; i <= SEND(self->embedded_string, length); i++) {        
        char c = SEND(self->embedded_string, charAt, i);
        switch (c) {
            case '~':
                SEND(self->plain_string, concatChars, "tilde"); 
                break;
            case '&':
                SEND(self->plain_string, concatChars, "and"); 
                break;
            case '|':
                SEND(self->plain_string, concatChars, "bar");
                break;
            case '*':
                SEND(self->plain_string, concatChars, "star");
                break;
            case '/':
                SEND(self->plain_string, concatChars, "slash");
                break;
            case '@':
                SEND(self->plain_string, concatChars, "at");
                break;
            case '+':
                SEND(self->plain_string, concatChars, "plus");
                break;
            case '-':
                SEND(self->plain_string, concatChars, "minus");
                break;
            case '=':
                SEND(self->plain_string, concatChars, "equal");
                break;     
            case '>':
                SEND(self->plain_string, concatChars, "greaterthan");
                break;
            case '<':
                SEND(self->plain_string, concatChars, "lowerthan");
                break;
            case ',':
                SEND(self->plain_string, concatChars, "comma");
                break;
            case '%':
                SEND(self->plain_string, concatChars, "percent");
                break;
            case '\\':
                SEND(self->plain_string, concatChars, "backslash");
                break;
            case ':':
                SEND(self->plain_string, concatChar, '_');
                break;
        #ifdef EXPERIMENTAL
            case ' ':
                SEND(self->plain_string, concatChar, '_');
                break;
        #endif EXPERIMENTAL
            default:
                SEND(self->plain_string, concatChar, c);
                break;
        }
    }
    return self->plain_string;
}


void _VMSymbol_mark_references(void* _self) {
    pVMSymbol self = (pVMSymbol) _self;
    SUPER(VMObject, self, mark_references);
}


//
// The VTABLE-function
//
//
static VTABLE(VMSymbol) _VMSymbol_vtable;
bool VMSymbol_vtable_inited = false;

VTABLE(VMSymbol)* VMSymbol_vtable(void) {
    if(! VMSymbol_vtable_inited) {
        *((VTABLE(VMObject)*)&_VMSymbol_vtable) = *VMObject_vtable();
        _VMSymbol_vtable.init             = METHOD(VMSymbol, init);        
        _VMSymbol_vtable.free             = METHOD(VMSymbol, free);
        _VMSymbol_vtable.get_string       = METHOD(VMSymbol, get_string);
        _VMSymbol_vtable.get_plain_string = METHOD(VMSymbol, get_plain_string);
        _VMSymbol_vtable.set_string       = METHOD(VMSymbol, set_string);
        
        _VMSymbol_vtable.mark_references = 
            METHOD(VMSymbol, mark_references);

        VMSymbol_vtable_inited = true;
    }
    return &_VMSymbol_vtable;
}

