package de.sm.metrics;

import java.io.File;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Examines a given folder and collect the files to be analyzed.
 * @author Stefan Marr
 */
public class FileSelector {
	
	/**
	 * Root directory to be analyzed.
	 */
	private File baseFolder;
	
	private List<File> vmadlFiles;
	private List<File> cFiles;
	private List<File> ahFiles;
	private List<File> hFiles;
	
	private Map<String, List<File>> mapExtToList;

	/**
	 * Initialize FileSelector for further processing.
	 * @param startDirectory root dir for source base to be analyzed
	 */
	public FileSelector(final String startDirectory) {
		baseFolder = new File(startDirectory);
		
		vmadlFiles = new ArrayList<File>();
		cFiles = new ArrayList<File>();
		ahFiles = new ArrayList<File>();
		hFiles = new ArrayList<File>();
		
		mapExtToList = new HashMap<String, List<File>>(); 
		mapExtToList.put(".vmadl", vmadlFiles);
		mapExtToList.put(".c", cFiles);
		mapExtToList.put(".h", hFiles);
		mapExtToList.put(".ah", ahFiles);
	}

	/**
	 * Examines recursively the baseFolder and tells the different
	 * file types apart to be analyzed by the right parser later. 
	 */
	public final void collect() {
		if (!baseFolder.isDirectory()) { 
			throw new InvalidParameterException(
					"Given path did not point to a directory");
		}
		
		collectFiles(baseFolder.listFiles());
	}

	/**
	 * Does the real work for collect().
	 * @param files list of files to be analyzed
	 */
	private void collectFiles(final File[] files) {
		if (files == null) { return; }
		
		for (File file : files) {
			if (file.isDirectory()) {
				collectFiles(file.listFiles());
			} else {
				selectByName(file);
			}
		}
	}

	/**
	 * Put file to the right list.
	 * @param file to be put in a list
	 */
	private void selectByName(final File file) {
		String name = file.getName();
		
		int pos = name.lastIndexOf('.');
		if (pos < 0) { return; }	// not found
		
		String ext = name.substring(pos);
		if (mapExtToList.containsKey(ext)) {
			mapExtToList.get(ext).add(file);
		}
	}
	
	public List<File> getFilesForLineCount() {
		List<File> list = new ArrayList<File>();
		list.addAll(vmadlFiles);
		list.addAll(cFiles);
		list.addAll(hFiles);
		list.addAll(ahFiles);
		
		return list;
	}

	public List<File> getFilesForCStats() {
		List<File> list = new ArrayList<File>();
		list.addAll(cFiles);
		list.addAll(hFiles);
		
		return list;
	}

	public List<File> getFilesForAspectCStats() {
		List<File> list = new ArrayList<File>();
		list.addAll(ahFiles);
		
		return list;
	}

	public List<File> getFilesForVmadlStats() {
		List<File> list = new ArrayList<File>();
		list.addAll(vmadlFiles);
		
		return list;
	}

}
