/**
 * 
 */
package de.sm.metrics;

import java.io.File;
import java.io.IOException;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

import de.sm.vmadl.CombinationRegistry;
import de.sm.vmadl.ServiceRegistry;
import de.sm.vmadl.VmadlLexer;
import de.sm.vmadl.VmadlParser;
import de.sm.vmadl.VmadlParser.architecture_return;
import de.sm.vmadl.elements.Combination;
import de.sm.vmadl.elements.ServiceInterface;
import de.sm.vmadl.elements.ServiceModule;

/**
 * @author gron
 *
 */
public class VmadlMetrics extends Metrics {

	private architecture_return parseResult;
	private int declarationCount;
	private int pointcutCount;
	private int typeCount;
	private int adviceCount;
	private int functionCount;
	private int preprocessorCount;
	private int variablesCount;
	
	/**
	 * @param file
	 * @throws RecognitionException 
	 * @throws IOException 
	 */
	public VmadlMetrics(File file) throws IOException, RecognitionException {
		super(file);
		parseFile();
	}
	
	private void parseFile() throws IOException, RecognitionException {
		ServiceRegistry serviceReg = new ServiceRegistry();
		CombinationRegistry combReg = new CombinationRegistry();
		
		CharStream input = new ANTLRFileStream(file.getAbsolutePath());
		VmadlLexer lexer = new VmadlLexer(input);
		
        CommonTokenStream tokens = new  CommonTokenStream(lexer);
		VmadlParser parser = new VmadlParser(tokens);
		parser.setParsedFile(file);
	    
		parseResult = parser.architecture(serviceReg, combReg);
		
		extractMetricsFromResult();
	}
	
	private void extractMetricsFromResult() {
		for (ServiceModule m : parseResult.modules) {
			if (m.initialized) {
				analyseService(m);
			}
		}
		
		for (Combination comb : parseResult.combinations) {
			adviceCount += comb.advices.size();
			adviceCount += comb.shutdownAdvices.size();
			adviceCount += comb.startupAdvices.size();
			functionCount += comb.functions.size();
			preprocessorCount += comb.preprocessor.size();
			variablesCount += comb.variables.size();
		}
	}
	
	private void analyseService(ServiceInterface service) {
		declarationCount += service.declarations.size();
		pointcutCount += service.pointcuts.size();
		pointcutCount += service.shutdownPointcuts.size();
		pointcutCount += service.startupPointcuts.size();
		typeCount += service.types.size();

		for (ServiceInterface s : service.parts.values()) {
			analyseService(s);
		}
	}
	
	public int getDeclarationCount() {
		return declarationCount;
	}

	public int getPointcutCount() {
		return pointcutCount;
	}
	
	public int getAdviceCount() {
		return adviceCount;
	}
	
	public int getTypeCount() {
		return typeCount;
	}
	
	/*public int getFunctionDefinitionCount() {
		
	}*/
	
	public int getCombineCount() {
		return parseResult.combinations.size();
	}
	
	public int getFunctionCount() {
		return functionCount;
	}
	
	public int getPreprocessorCount() {
		return preprocessorCount;
	}
	
	public int getVariableCount() {
		return variablesCount;
	}
}
