package de.sm.odl;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

import de.sm.csimple.CSimpleLexer;
import de.sm.csimple.CSimpleParser;

/**
 * The FunctionExtractor parses all .c files in a given directory.
 * Found function implementations are registered and it can be checked 
 * whether a given name is an implemented function. 
 *  
 * @author Stefan Marr
 */
public class FunctionExtractor {
	
	/**
	 * A set with names of implemented functions.
	 */
	private Set<String> functions = new HashSet<String>();
	
	/**
	 * Check whether there is a function implemented with the given name.
	 * @param name of function
	 * @return true if implementation is available
	 */
	public final boolean isFunctionDefined(final String name) {
		return functions.contains(name.intern());
	}
	
	/**
	 * Recursively extract function implementations from directory.   
	 * @param rootDir to be parsed
	 */
	public final void analyze(final File rootDir) {
		recurseDir(rootDir);
	}

	/**
	 * Recursively parse files.
	 * @param dir directory to be parsed
	 */
	private void recurseDir(final File dir) {
		File[] files = dir.listFiles();
		
		for (File file : files) {
			if (file.isDirectory()) {
				// TODO: remove this dead code and rename function
				recurseDir(file);
			} else {
				if (file.getName().endsWith(".c")) {
					System.err.println("Parse file: " + file.getName());
					parseFile(file);
				}
			}
		}
	}

	/**
	 * Parse given file and extract defined functions.
	 * @param file to be parsed
	 */
	private void parseFile(final File file) {
		CharStream input;
		
		try {
			input = new ANTLRFileStream(file.getAbsolutePath());
			
			CSimpleLexer lexer = new CSimpleLexer(input);
			CommonTokenStream tokens = new  CommonTokenStream(lexer);
			CSimpleParser parser = new CSimpleParser(tokens);
			parser.setParsedFile(file);
			
		    parser.definitions();
		    List<String> funcs = parser.getFuntions();
		    functions.addAll(funcs);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
