package de.sm.odl;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.sm.misc.MakeFormatter;
import de.sm.odl.elements.Field;
import de.sm.odl.elements.Function;
import de.sm.odl.elements.FunctionParam;
import de.sm.odl.elements.TraitType;
import de.sm.odl.elements.Type;
import de.sm.vmadl.elements.ServiceModule;

/**
 * Compiler for the Object Definition Language used to describe classes
 * for CSOM.
 * @author Stefan Marr
 */
public class Odl {
	
	private String frameTpl;
	private String typedefTpl;
	private String structNameTpl;
	private String pointerNameTpl;
	private String vtableTpl;
	private String vtableFuncPtrTpl;
	private String objectFormatTpl;
	private String objectFieldTpl;
	private String vtableInitForwardDeclTpl;
	private String vtableInitFuncTpl;
	private String vtableFuncInitTpl;
	private String methodNameTpl;
	private String assignTraitTpl;
	private String implementationHeader;
	private String vtableFuncForwardTpl;
	
	private Logger log;
	
	private TypeFlattener flattener;
	private FunctionExtractor functions;
	private TypeRegistry reg;
	private File sourceFolder;
	
	private class UsedFunction {
		public UsedFunction(final Type type, final Function func) {
			this.type = type;
			this.func = func;
		}
		public Type type;
		public Function func;
	}
	
	private List<UsedFunction> usedFunctions = new ArrayList<UsedFunction>();
	
	public static Odl getInstance() {
		TypeRegistry reg = new TypeRegistry();
		Odl odl = new Odl(reg, new TypeFlattener(reg), new FunctionExtractor());
		
		return odl;
	}
	
	public final TypeRegistry getTypeRegistry() {
		return reg;
	}
	
	public Odl(final TypeRegistry reg, 
			final TypeFlattener flattener, 
			final FunctionExtractor funcExtractor) {
		log = Logger.getLogger("Odl");
		ConsoleHandler ch = new ConsoleHandler();
		ch.setFormatter(new MakeFormatter());
		log.addHandler(ch);
		
		initTemplates();
		this.reg = reg;
		functions = funcExtractor;
		this.flattener = flattener;
	}
	
	public final void setSourceFolder(final File sourceFolder) {
		this.sourceFolder = sourceFolder;
	}
	
	public final void analyze(final File sourceFolder) {
		functions.analyze(sourceFolder);
	}
	
	/**
	 * Initialize templates.
	 */
	private void initTemplates() {
		// %1 - Include Guard
		// %2 - Includes
		// %3 - content
		frameTpl = "#ifndef %1$s\n#define %1$s\n/**\n * This files is"
			+ " generated, do not modify, edit the VMADL/ODL file directly"
			+ " instead.\n"
			+ " */\n\n"
			+ "%2$s\n\n"
			+ "%3$s\n"
			+ "\n#endif /* %1$s */\n";
		
		// %1 Struct Name
		// %2 Type Name
		// %3 PointerName
		typedefTpl = "typedef %1$s %2$s, *%3$s;\n";
		
		// %1 Type name
		structNameTpl = "struct _%1$s";
		
		// %1 Type name
		pointerNameTpl = "p%1$s";
		
		// %1 Type name
		// %2 function pointers
		vtableTpl = "struct _%1$s_vtable_layout {\n"
			+ "\tvoid* _ttable;\n"
			+ "%2$s"
			+ "};\n\n";
		
		// %1 return type
		// %2 function name
		// %3 params including leading , for separation
		vtableFuncPtrTpl = "%1$s (*%2$s)(void* _self%3$s);\n";
		
		// %1 return type
		// %2 type name
		// %3 function name
		// %4 param list
		vtableFuncForwardTpl = "%1$s _%2$s_%3$s(void* _self%4$s);\n";
		
		// %1 struct name of type
		// %2 type name
		// %3 object fields
		objectFormatTpl = "%1$s {\n"
		    + "\tVTABLE(%2$s)* _vtable;\n"
		    + "%3$s"
		    + "};\n\n";
		
		// %1 type
		// %2 name
		objectFieldTpl = "%1$s %2$s;\n";
		
		// %1 type name
		vtableInitForwardDeclTpl = "VTABLE(%1$s)* %1$s_vtable(void);\n";
		
		// %1 type name
		// %2 function initializations
		vtableInitFuncTpl = "VTABLE(%1$s)* %1$s_vtable(void) {\n"
			+ "\tstatic VTABLE(%1$s) vtable;\n"
			+ "\tstatic bool inited = false;\n"
			+ "\tif (!inited) {\n"
			+ "\t\tvtable._ttable = NULL;\n"
			+ "%2$s\n"
			+ "\t\tinited = true;\n"
			+ "\t}\n"
			+ "\treturn &vtable;\n"
			+ "}\n";
		
		// %1 type name
		// %2 function name
		vtableFuncInitTpl = "vtable.%2$s = METHOD(%1$s, %2$s);\n";

		// %1 type name
		// %2 method name
		methodNameTpl = "_%1$s_%2$s";
		
		// %1 trait name
		assignTraitTpl = "vtable._ttable = %1$s_vtable();\n";
		
		implementationHeader = "/** This file is generated, please do not"
			+ " touch! **/\n\n";
	}
	
	private Type resolveDefiningClass(final Type type, final String functionName) {
		String fullName = String.format(methodNameTpl, type.getName(), functionName);
		
		if (functions.isFunctionDefined(fullName)) {
			return type;
		} else {
			for (TraitType trait : type.getTraits()) {
				Type result = resolveDefiningClass(trait, functionName);
				if (result != null) {
					return result;
				}
			}
			if (type.getBaseclass() != null) {
				return resolveDefiningClass(type.getBaseclass(), functionName);
			}
		}
		return null;
	}
	
	private String generateVtableFunction(final Type type) {
		StringBuilder inits = new StringBuilder();
		
		for (TraitType trait : type.getTraits()) {
			inits.append("\n\t\t");
			inits.append(String.format(assignTraitTpl, trait.getName()));
		}
		
		appendFunctions(type, inits, type);
		
		return String.format(vtableInitFuncTpl, type.getName(), inits.toString());
	}

	private void appendFunctions(final Type type, final StringBuilder inits, final Type current) {
		for (Function func : current.getFunctions()) {
			Type definingClass = resolveDefiningClass(type, func.name);
			if (definingClass == null) {
				log.log(Level.WARNING, "For " + type + " a function " + func.name + " has been declared but not implemented.");
				continue;
			}
			inits.append("\t\t");
			inits.append(String.format(vtableFuncInitTpl, definingClass.getName(), func.name));
			
			usedFunctions.add(new UsedFunction(definingClass, func));
		}
	}
	
	
	private String generateVtableInitForwardDeclaration(final Type type) {
		return String.format(vtableInitForwardDeclTpl, type.getName());
	}
	
	private String generateObjectFormat(final Type type) {
		StringBuilder fields = new StringBuilder();
		
		generateObjectFields(type, fields);
		
		return String.format(objectFormatTpl, String.format(structNameTpl, type.getName()), type.getName(), fields.toString());
	}
	
	private void generateObjectFields(final Type type, final StringBuilder fields) {
		for (Field field : type.getFields()) {
			fields.append("\t");
			fields.append(String.format(objectFieldTpl, field.type.getCompleteName(), field.name));
		}
	}

	private String generateVtable(final Type type) {
		StringBuilder ptrs = new StringBuilder();
		
		generateVtableFunctPtrs(type, ptrs);
		
		return String.format(vtableTpl, type.getName(), ptrs.toString()); 
	}
	
	private void generateVtableFunctPtrs(final Type type, final StringBuilder ptrs) {
		for (Function func : type.getFunctions()) {
			ptrs.append("\t");
			ptrs.append(generateVtableFuncPtr(func));
		}
	}

	private String generateVtableFuncPtr(final Function func) {
		String paramList = parameterList(func);
		return String.format(vtableFuncPtrTpl, func.returnType.getCompleteName(), func.name, paramList);
	}

	private String parameterList(final Function func) {
		StringBuilder sb = new StringBuilder();
		for (FunctionParam param : func.getParams()) {
			sb.append(", ");
			sb.append(param.type.getCompleteName());
			if (param.name != null) {
				sb.append(" ");
				sb.append(param.name);
			}
		}
		return sb.toString();
	}
	
	private String getTypedef(final String typeName) {
		String struct = String.format(structNameTpl, typeName);
		String ptr = String.format(pointerNameTpl, typeName);
		return String.format(typedefTpl, struct, typeName, ptr);
	}
	
	private String generateTypedefs(final ServiceModule module) {
		StringBuilder sb = new StringBuilder();
		for (Type type : flattener.getAllFor(module)) {
			sb.append(getTypedef(type.getName()));
		}
		return sb.toString();
	}
	
	/**
	 * Generate the forward declarations header file for the given module.
	 * @param targetFile file to write header to
	 * @param module for which headers have to be generated
	 */
	public final void generateForwardHeader(final File targetFile,
			final ServiceModule module) {
		StringBuilder content = new StringBuilder();
		
		content.append("// Forward declarations of generated types\n");
		content.append(generateTypedefs(module));
		content.append("\n");
		
		writeHeaderFile(targetFile, "", content);
	}
	
	/**
	 * Generate the type definition header file for the given module.
	 * @param targetFile
	 * @param module
	 */
	public final void generateDefinitionHeader(final File targetFile, final ServiceModule module, final String includes) {
		StringBuilder content = new StringBuilder();
		content.append("// Type structures and VTables\n\n");
		
		for (Type type : flattener.getAllFor(module)) {
			content.append(generateVtable(type));
			content.append("\n\n");
			
			content.append(generateObjectFormat(type));
			content.append("\n\n");
			
			
			content.append(generateVtableInitForwardDeclaration(type));
			content.append("\n");
		}
		
		writeHeaderFile(targetFile, includes, content);
	}

	private void writeHeaderFile(final File targetFile, final String includes,
			final StringBuilder content) {
		String includeGuard = "_" + targetFile.getName().replace(".", "_") + "_";
		FileWriter fw;
		try {
			fw = new FileWriter(targetFile);
			fw.append(String.format(frameTpl, 
					includeGuard.toUpperCase(),
					includes, content.toString()));
			fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public final void generateImplementation(final File targetFile, final String includes) {
		StringBuilder content = new StringBuilder();
		StringBuilder vtableInits = new StringBuilder();
		content.append(implementationHeader);
		content.append(includes);
		
		for (Type type : flattener.getAll()) {
			vtableInits.append(generateVtableFunction(type));
			vtableInits.append("\n\n");
		}
		
		for (UsedFunction uf : usedFunctions) {
			content.append(String.format(vtableFuncForwardTpl, 
					uf.func.returnType.getCompleteName(),
					uf.type.getName(),
					uf.func.name,
					parameterList(uf.func)));	
		}
		
		content.append("\n\n");
		content.append(vtableInits);
		
		try {
			FileWriter fw = new FileWriter(targetFile);
			fw.append(content.toString());
			fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}

	public final void prepareCompilation() {
		flattener.doFlattening();
	}
}
