package de.sm.odl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map; 

import de.sm.odl.elements.Field;
import de.sm.odl.elements.Function;
import de.sm.odl.elements.TraitType;
import de.sm.odl.elements.Type;
import de.sm.vmadl.elements.ServiceModule;

public class TypeFlattener {
	
	private Map<String, Type> types;
	private TypeRegistry typeReg;
	
	public TypeFlattener(TypeRegistry registry) {
		types = new HashMap<String, Type>();
		typeReg = registry;
	}
	
	public Type get(String name) {
		return types.get(name);
	}
	
	public Collection<Type> getAll() {
		return types.values();
	}
	
	public Collection<Type> getAllFor(ServiceModule module) {
		Collection<Type> choosen = new ArrayList<Type>();
		for (Type t : types.values()) {
			if (t.getModule() == module) {
				choosen.add(t);
			}
		}
		return choosen;
	}
	
	public void doFlattening() {
		for (Type type : typeReg.getAll()) {
			flattenType(type);
		}
	}

	private void flattenType(Type type) {
		if (types.containsKey(type.getName())) return;
		
		Type base = null;
		if (type.getBaseclass() != null) {
			if (!types.containsKey(type.getName())) {
				flattenType(type.getBaseclass());
			}
			base = types.get(type.getBaseclass().getName());
		}
		
		for (TraitType trait : type.getTraits()) {
			flattenType(trait);
		}
		
		// propagate down traits relationship
		if (base != null) {
			type.getTraits().addAll(0, base.getTraits());
		}
		
		try {
			Type flattenedType = (Type)type.clone();
			handleFieldConstraints(flattenedType, base);
			handleFunctions(flattenedType, base);
			
			types.put(flattenedType.getName(), flattenedType);
		} catch (CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void handleFunctions(Type flattenedType, Type base) {
		//List<Function> functions = flattenedType.getFunctions();
		
		Map<String, Function> functions = new LinkedHashMap<String, Function>();
		
		// Add functions from base class
		if (base != null) {
			for (Function func : base.getFunctions()) {
				if (!functions.containsKey(func.name)) {
					functions.put(func.name, func);
				}
			}
		}
		
		// Add functions from traits
		for (TraitType trait : flattenedType.getTraits()) {
			for (Function func : trait.getFunctions()) {
				if (!functions.containsKey(func.name)) {
					functions.put(func.name, func);
				}
			}
		}
		
		// Add functions from current class
		for (Function func : flattenedType.getFunctions()) {
			if (!functions.containsKey(func.name)) {
				functions.put(func.name, func);
			}
		}
		
		List<Function> result = flattenedType.getFunctions();
		result.clear();
		result.addAll(functions.values());
	}

	private void handleFieldConstraints(Type flattenedType, Type base) {
		List<Field> fields = flattenedType.getFields();
		
		if (base != null) {
			fields.addAll(0, base.getFields());
		}
		
		// Remark: Trait Fields are left out intentionally
		
		for (int i = 0; i < fields.size(); i++) {
			Field field = fields.get(i);
			if ("before".equals(field.positionPredicate)) {
				int pos = getIndexByName(fields, field.positionField);
				if (i > pos && pos != -1) {  // predicate violated
					fields.remove(i);
					fields.add(pos, field);
				}
			}
		}
	}

	private int getIndexByName(List<Field> fields, String positionField) {
		for (int i = 0; i < fields.size(); i++) {
			if (positionField.equals(fields.get(i).name)) {
				return i;
			}
		}
		return -1;
	}
}
