package de.sm.vmadl;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Formatter;
import java.util.List;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

import de.sm.vmadl.VmadlParser.architecture_return;
import de.sm.vmadl.elements.OptionalRelation;
import de.sm.vmadl.elements.ServiceInterface;
import de.sm.vmadl.elements.ServiceModule;

public class Vmadl {
	
	private ServiceRegistry reg;
	private String headerFile;
	private String aspectFile;
	private File outputBase;
	
	public Vmadl() {
		reg = new ServiceRegistry();
		initTemplates();
	}
	
	private void initTemplates() {
		// %1 - Include Guard
		// %2 - Name of the vmadl file
		// %3 - Content
		headerFile = "#ifndef %1$s\n#define %1$s\n/**\n * This files is generated, do not modify, edit the VMADL file directly instead.\n"
			+" * Generated from: %2$s\n"
			+" */\n\n"
			+"%3$s\n"
			+"\n#endif /* %1$s */\n";
		
		// %1 - Include guard
		// %2 - Name of the vmadl file
		// %3 - Includes
		// %4 - Aspect name
		// %5 - Aspect content
		aspectFile = "#ifndef %1$s\n#define %1$s\n/**\n * This files is generated, do not modify, edit the VMADL file directly instead.\n"
			+" * Generated from: %2$s\n"
			+" */\n\n"
			+"\n%3$s\n\n"
			+"aspect %4$s {\n"
			+"%5$s\n\n"
			+"};\n"
			+"\n#endif /* %1$s */\n";
	}
	
	private void writeHeaderFile(File targetFile, String includeGuard, String vmadlFilename, String content) {
		FileWriter fw;
		try {
			//StringBuilder sb = new StringBuilder();
			Formatter formatter = new Formatter(targetFile);
			formatter.format(headerFile, includeGuard, vmadlFilename, content);
			formatter.close();
			//fw = new FileWriter(targetFile);
			//fw.write(String.format(headerFile, includeGuard, vmadlFilename, content));
			//fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private void writeAspectFile(File targetFile, String includeGuard, String vmadlFilename, String includes, String aspectName, String content) {
		FileWriter fw;
		try {
			fw = new FileWriter(targetFile);
			fw.write(String.format(aspectFile, includeGuard, vmadlFilename, includes, aspectName, content));
			fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Loads the VM Architecture description from the given file.
	 * @return List<VmadlDesc>
	 */
	public List<ServiceModule> loadFromFile(File file) {
		CharStream input;
		try {
			input = new ANTLRFileStream(file.getAbsolutePath());
			VmadlLexer lexer = new VmadlLexer(input);
			
	        CommonTokenStream tokens = new  CommonTokenStream(lexer);
			VmadlParser parser = new VmadlParser(tokens);
			
		    
		    architecture_return arch = parser.architecture(reg);
		    
		    for (ServiceModule module : arch.modules) {
		    	module.file = file;
		    }
		    
		    return arch.modules;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	
	public void checkConsistency() {
		// verify internal consistency of VMADL files
		// all used pointcuts in advices are only allowed to use function names or pointcuts defined
		// in other vmadl files, and therefore the right require statements have to be given.
		// Same could be proven for startup and shutdown, advice are only allowed to use pointcuts available
		// at this time
	}
	
	public void generateImplementation(String outputDir) {
		outputBase = new File(outputDir);
		if (!outputBase.exists() && !outputBase.mkdirs()) {
			throw new RuntimeException("Output directory ("+ outputDir +") could not be created.");
		}
		
		// At first resolve optional stuff (OptionalRelation)
		resolveOptionalRelations();
		
		Collection<ServiceModule> services = reg.getServices();
		for (ServiceModule module : services) {
			if (module.initialized) {
				generateHeader(module);
				if (hasAspects(module)) { generateAspect(module); }
			}
		}
	}
	
	private void resolveOptionalRelations() {
		Collection<ServiceModule> services = reg.getServices();
		for (ServiceModule module : services) {
			for (OptionalRelation relation : module.relations) {
				if (relation.service.initialized) {
					module.advices.addAll(relation.advices);
					module.declarations.addAll(relation.declarations);
					module.pointcuts.addAll(relation.pointcuts);
					module.shutdownAdvices.addAll(relation.shutdownAdvices);
					module.shutdownPointcuts.addAll(relation.shutdownPointcuts);
					module.startupAdvices.addAll(relation.startupAdvices);
					module.startupPointcuts.addAll(relation.startupPointcuts);
					module.requirements.addAll(relation.requirements);
					module.requirements.add(relation.service);
				}
			}
		}
	}

	private boolean hasAspects(ServiceInterface module) {
		return !module.advices.isEmpty() || !module.pointcuts.isEmpty()
			|| !module.shutdownAdvices.isEmpty() || !module.shutdownPointcuts.isEmpty()
			|| !module.startupAdvices.isEmpty() || !module.startupPointcuts.isEmpty();
	}
	
	private void generateAspect(ServiceModule module) {
		StringBuilder content = new StringBuilder();
		StringBuilder includes = new StringBuilder();
		
		generateAspectIncludes(module, includes);
		
		content.append("\t/*** Pointcuts ***/\n");
		for (String pointcut : module.pointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
		
		content.append("\t/*** Advices ***/\n");
		for (String advice : module.advices) {
			content.append("#line 55 \"");
			content.append(module.file.getAbsolutePath());
			content.append("\"\n");
			content.append(advice);
			content.append("\n\n");
		}

		content.append("\t/*** Pointcuts only available at startup time ***/\n");
		for (String pointcut : module.startupPointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
		
		content.append("\t/*** Advices only active at startup time ***/\n");
		content.append("\t/* REMARK: propably there should be a guard to ensure this, automatically generated. */");
		for (String advice : module.startupAdvices) {
			content.append(advice);
			content.append("\n\n");
		}
		
		content.append("\t/*** Pointcuts only available at shutdown time ***/\n");
		for (String pointcut : module.shutdownPointcuts) {
			content.append(pointcut);
			content.append("\n\n");
		}
		
		content.append("\t/*** Advices only active at shutdown time ***/\n");
		content.append("\t/* REMARK: propably there should be a guard to ensure this, automatically generated. */");
		for (String advice : module.shutdownAdvices) {
			content.append(advice);
			content.append("\n\n");
		}
		
		File outputFile = new File(outputBase, "module." + module.name + ".ah");
		String guard = getGuardFromName(module.name, "ah");
		writeAspectFile(outputFile, guard, module.file.getName(), includes.toString(), module.name, content.toString());
	}
	
	private void generateAspectIncludes(ServiceModule module, StringBuilder includes) {
		// Self include (of the own header file)
		includes.append("#include <module.");
		includes.append(module.name);
		includes.append(">\n");
		
		generateIncludes(module, includes);
		for (ServiceModule req : module.requirements) {
			if (hasAspects(req)) {
				includes.append("#include <module.");
				includes.append(req.name);
				includes.append(".ah>\n");
			}
		}
	}

	private void generateIncludes(ServiceModule module, StringBuilder includes) {
		for (ServiceModule req : module.requirements) {
			includes.append("#include <module.");
			includes.append(req.name);
			includes.append(">\n");
		}
	}
	
	

	private void generateHeader(ServiceModule module) {
		File outputFile = new File(outputBase, "module." + module.name);
		String guard = getGuardFromName(module.name, "h");
		StringBuilder content = new StringBuilder();
				
		generateIncludes(module, content);
		
		for (String decl : module.declarations) {
			content.append(decl);
			content.append("\n");
		}
		
		writeHeaderFile(outputFile, guard, module.file.getName(), content.toString());
	}

	private String getGuardFromName(String name, String type) {
		return name.toUpperCase() + "_" + type.toUpperCase() + "_";
	}

}
