
#include <stdio.h>
#include <unistd.h>

#include "Scheduler.h"

#include <vm/Universe.h>

#include <interpreter/bytecodes.h>
#include <interpreter/Interpreter.h>

#include <vmobjects/VMMethod.h>
#include <vmobjects/VMArray.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMSymbol.h>

#include <memory/memory.h>




typedef struct ThreadToFrameAssociation threadToFrameAssociation;

//Association between Smalltalk Thread Object and VM frame
struct ThreadToFrameAssociation {
	pVMObject thread;
	pVMFrame frame;
	threadToFrameAssociation* next;
};

//thread list
threadToFrameAssociation* frames = NULL;
//currently running thread
threadToFrameAssociation* currentThread = NULL;
//scheduler thread and frame for preemption
threadToFrameAssociation* schedulerThread = NULL;

bool firstRun = true;

void Scheduler_run_thread(pVMObject thread, pVMFrame frame) {
	//save the scheduler thread only at the first run
	if (firstRun) {
		firstRun = false;
		schedulerThread = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		schedulerThread->thread = thread;
		schedulerThread->frame = frame;
		schedulerThread->next = NULL;
		Interpreter_set_scheduler_frame(frame);
	}
		 
	if (currentThread != NULL) {
		currentThread->frame = frame;
	}
		
	threadToFrameAssociation* frameAss = frames;
	while (frameAss->thread != thread) {
		if (frameAss->next == NULL) {
			return;
		}
		frameAss = frameAss->next;
	}
	
	currentThread = frameAss;
	Interpreter_set_current_frame(frameAss->frame);
}

void  Scheduler_register_thread_frame(pVMObject arg) {
	pVMClass cl = SEND(arg, get_class);

	//create surrounding frame - bootstrap frame
	pVMMethod bootstrap_method = Interpreter_create_bootstrap_method(cl);

	pVMFrame bootstrap_frame = VMFrame_new_with(NULL, (pVMMethod)bootstrap_method);
    SEND(bootstrap_frame, push, arg);

	//and set the run method as the method to invoke
	pVMInvokable runMethod = (pVMInvokable)SEND(cl, lookup_invokable, Universe_symbol_for_chars("run"));
	pVMFrame newFrame = VMFrame_new_with(bootstrap_frame, (pVMMethod)runMethod);
	SEND(newFrame, set_argument, 0, 0, arg);

	//create first linked list entry if not created yet
	if (frames == NULL) {
		frames = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		frames->thread = arg;
		frames->frame = newFrame;
		frames->next = NULL;
	}
	//or append the new association to the list
	else {
		threadToFrameAssociation* frame = frames->next;
		threadToFrameAssociation* prevFrame = frames;
		while (frame != NULL) {
			prevFrame = frame;
			frame = frame->next;
		}
		frame = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		frame->thread = arg;
		frame->frame = newFrame;
		frame->next = NULL;
		prevFrame->next = frame;
	}
}

void Scheduler_unregister_thread_frame(pVMObject thread) {
	threadToFrameAssociation* threadFrame = frames;
	threadToFrameAssociation* prevFrame = NULL;
	while (threadFrame != NULL && threadFrame->thread != thread) {
		prevFrame = threadFrame;
		threadFrame = threadFrame->next;
	}
	if (threadFrame != NULL) {
		if (prevFrame != NULL)
			prevFrame->next = threadFrame->next;
		threadFrame = NULL;
	}
}

void Scheduler_run_scheduler_thread() {
	if (currentThread != NULL) {
		Interpreter_get_frame(&currentThread->frame);
	}

	currentThread = schedulerThread;
	Interpreter_set_current_frame(schedulerThread->frame);
}

void Scheduler_sleep(long duration) {
	sleep(duration);
}
