/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <stdio.h>
#include <string.h>

#include "Disassembler.h"

#include <misc/debug.h>

#include <vm/Universe.h>

#include <interpreter/bytecodes.h>
#include <interpreter/Interpreter.h>

#include <vmobjects/VMArray.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMBlock.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMDouble.h>
#include <vmobjects/VMEvaluationPrimitive.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMMethod.h>
#include <vmobjects/VMObject.h>
#include <vmobjects/VMPrimitive.h>
#include <vmobjects/VMString.h>
#include <vmobjects/VMSymbol.h>
#include <vmobjects/Signature.h>

/** 
 * Dispatch an object to its content and write out
 */
static inline void _Disassembler_dispatch(pVMObject o) {
    //dispatch
    // can't switch() objects, so:
    if(!o) return; // NULL isn't interesting.
    else if(o == nil_object)
        debug_print("{Nil}");
    else if(o == true_object)
        debug_print("{True}");
    else if(o == false_object)
        debug_print("{False}"); 
    else if((pVMClass)o == system_class)
        debug_print("{System Class object}");
    else if((pVMClass)o == block_class)
        debug_print("{Block Class object}");
    else if(o == Universe_get_global(Universe_symbol_for_chars("system")))
        debug_print("{System}");
    else {
        pVMClass c = SEND(o, get_class);
        if(c == string_class) {
            pString s = SEND((pVMString)o, get_embedded_string);
            debug_print("\"%s\"", SEND(s, chars));
        } else if(c == double_class)
            debug_print("%g", SEND((pVMDouble)o, get_embedded_double));
        else if(c == biginteger_class)
            debug_print("%lld", SEND((pVMBigInteger)o,
                                     get_embedded_biginteger));
        else if(c == integer_class)
            debug_print("%d", SEND((pVMInteger)o, get_embedded_integer));
        else if(c == symbol_class) {
            pString s = SEND((pVMSymbol)o, get_string);
            debug_print("#%s", SEND(s, chars));
        } else
            debug_print("address: %p", (void*)o);
    }
}

/**
 * Dump a class and all subsequent methods.
 */
void Disassembler_dump(pVMClass clazz) {
    for(int i = 0; i < SEND(clazz, get_number_of_instance_invokables); i++) {
        pVMObject inv = SEND(clazz, get_instance_invokable, i);
        // output header and skip if the Invokable is a Primitive
        pVMSymbol sig = TSEND(VMInvokable, inv, get_signature);
        pString sig_s = SEND(sig, get_string);
        pVMSymbol cname = SEND(clazz, get_name);
        pString cname_s = SEND(cname, get_string);
        debug_dump("%s>>%s = ", SEND(cname_s, chars), SEND(sig_s, chars));
        if(TSEND(VMInvokable, inv, is_primitive)) {
            debug_print("<primitive>\n");
            continue;
        }
        // output actual method
        Disassembler_dump_method((pVMMethod)inv, "\t");
    }
}


/**
 * Bytecode Index Accessor macros
 */
#define BC_0 SEND(method, get_bytecode,bc_idx)
#define BC_1 SEND(method, get_bytecode,bc_idx+1)
#define BC_2 SEND(method, get_bytecode,bc_idx+2)


/**
 * Dump all Bytecode of a method.
 */
void Disassembler_dump_method(pVMMethod method, const char* indent) {
    debug_print("(\n");
    {   // output stack information
        int locals =SEND(method, get_number_of_locals); 
        int max_stack = SEND(method, get_maximum_number_of_stack_elements);  
        debug_dump("%s<%d locals, %d stack>\n", indent, locals, max_stack);
    }
    // output bytecodes
    bytecode_t bc_len;
    for(size_t bc_idx = 0; 
        bc_idx < SEND(method, get_number_of_bytecodes); 
        bytecodes_get_bytecode_length(
            SEND(method, get_bytecode, bc_idx), &bc_len), bc_idx += bc_len
    ) {
        // the bytecode.
    	bytecode_t bytecode = BC_0;
        // indent, bytecode index, bytecode mnemonic
        debug_dump("%s%4d:%s  ", indent, bc_idx,
            bytecodes_get_bytecode_name(bytecode));
        // parameters (if any)
        bytecode_t length;
        bytecodes_get_bytecode_length(bytecode, &length);
        if (length == 1) {
            debug_print("\n");
            continue;
        }
        switch(bytecode) {
            case BC_PUSH_LOCAL:
                debug_print("local: %d, context: %d\n", BC_1, BC_2); break;
            case BC_PUSH_ARGUMENT:
                debug_print("argument: %d, context %d\n", BC_1, BC_2); break;
            case BC_PUSH_FIELD:{
                pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
                pString s_name = SEND(name, get_string);
                debug_print("(index: %d) field: %s\n", BC_1,
                    SEND(s_name, chars));
                break;
            }
            case BC_PUSH_BLOCK: {
                char nindent[strlen(indent)+1+1];
                debug_print("block: (index: %d) ", BC_1);
                sprintf(nindent, "%s\t", indent);
                Disassembler_dump_method(
                    (pVMMethod)SEND(method, get_constant, bc_idx), nindent);
                break;
            }            
            case BC_PUSH_CONSTANT: {
                pVMObject constant = SEND(method, get_constant, bc_idx);
                pVMClass clazz = SEND(constant, get_class);
                pVMSymbol cname = SEND(clazz, get_name);
                pString s_cname = SEND(cname, get_string);
                debug_print("(index: %d) value: (%s) ", 
                            BC_1, SEND(s_cname, chars));
                _Disassembler_dispatch(constant); debug_print("\n");
                break;
            }
            case BC_PUSH_GLOBAL: {
                pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
                pString s_name = SEND(name, get_string);
                debug_print("(index: %d) value: %s\n", BC_1,
                            SEND(s_name, chars));
                break;
            }
            case BC_POP_LOCAL:
                debug_print("local: %d, context: %d\n", BC_1, BC_2);
                break;
            case BC_POP_ARGUMENT:
                debug_print("argument: %d, context: %d\n", BC_1, BC_2);
                break;
            case BC_POP_FIELD: {
                pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
                pString s_name = SEND(name, get_string);
                debug_print("(index: %d) field: %s\n", BC_1, SEND(s_name, chars));
                break;
            }
            case BC_SEND: {
                pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
                pString s_name = SEND(name, get_string);
                debug_print("(index: %d) signature: %s\n", BC_1,
                    SEND(s_name, chars));
                break;
            }
            case BC_SUPER_SEND: {
                pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
                pString s_name = SEND(name, get_string);
                debug_print("(index: %d) signature: %s\n", BC_1,
                    SEND(s_name, chars));
                break;
            }
            default:
                debug_print("<incorrect bytecode>\n");
                break;
        }
    }
    debug_dump("%s)\n", indent);
}

/**
 * Dump bytecode from the frame running
 */
void Disassembler_dump_bytecode(pVMFrame frame, pVMMethod method, int bc_idx) {
    static long long indentc = 0;
    static char      ikind   = '@';
    bytecode_t       bc      = BC_0;
    pVMClass         clazz   = TSEND(VMInvokable, method, get_holder);
    
    // Determine Context: Class or Block?
    if(IS_A(clazz, VMClass)) {
        pVMSymbol cname = SEND(clazz, get_name);
        pVMSymbol sig = TSEND(VMInvokable, method, get_signature);
        pString s_cname = SEND(cname, get_string);
        pString s_sig = SEND(sig, get_string);
        
        debug_trace("%20s>>%-20s% 10lld %c %04d: %s\t",
                    SEND(s_cname, chars), SEND(s_sig, chars),
                    indentc, ikind, bc_idx,
                    bytecodes_get_bytecode_name(bc));        
    } else {
        pVMSymbol sig = TSEND(VMInvokable, method, get_signature);
        pString s_sig = SEND(sig, get_string);
        debug_trace("%-42s% 10lld %c %04d: %s\t", 
                    SEND(s_sig, chars),
                    indentc, ikind, bc_idx,
                    bytecodes_get_bytecode_name(bc));
    }
    // reset send indicator
    if(ikind != '@') ikind = '@';
    
    switch(bc) {
        case BC_HALT: {
            debug_print("<halting>\n\n\n");
            break;
        }
        case BC_DUP: {
            pVMObject o = SEND(frame, get_stack_element, 0);
            if(o) {
                pVMClass c = SEND(o, get_class);
                pVMSymbol cname = SEND(c, get_name);
                pString s_cname = SEND(cname, get_string);
                debug_print("<to dup: (%s) ", SEND(s_cname, chars));
                //dispatch
                _Disassembler_dispatch(o);
            } else
                debug_print("<to dup: address: %p", (void*)o);
            debug_print(">\n");                        
            break;
        }
        case BC_PUSH_LOCAL: {
        	bytecode_t bc1 = BC_1, bc2 = BC_2;
            pVMObject o = SEND(frame, get_local, bc1, bc2);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_cname = SEND(cname, get_string);
            debug_print("local: %d, context: %d <(%s) ", 
                        BC_1, BC_2, SEND(s_cname, chars));
            //dispatch
            _Disassembler_dispatch(o);
            debug_print(">\n");                        
            break;
        }
        case BC_PUSH_ARGUMENT: {
        	bytecode_t bc1 = BC_1, bc2 = BC_2;
            pVMObject o = SEND(frame, get_argument, bc1, bc2);
            debug_print("argument: %d, context: %d", bc1, bc2);
            if (IS_A(clazz, VMClass)) {
                pVMClass c = SEND(o, get_class);
                pVMSymbol cname = SEND(c, get_name);
                pString s_cname = SEND(cname, get_string);
                debug_print("<(%s) ", SEND(s_cname, chars));
                //dispatch
                _Disassembler_dispatch(o);                
                debug_print(">");                        
            }            
            debug_print("\n");
            break;
        }
        case BC_PUSH_FIELD: {
            pVMFrame ctxt = SEND(frame, get_outer_context);
            pVMObject arg = SEND(ctxt, get_argument, 0, 0);
            pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
            int field_index = SEND((pVMObject)arg, get_field_index, name);
           
            pVMObject o = SEND(arg, get_field, field_index);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_name = SEND(name, get_string);
            pString s_cname = SEND(cname, get_string);

            debug_print("(index: %d) field: %s <(%s) ", BC_1,
                        SEND(s_name, chars), SEND(s_cname, chars));
            //dispatch
            _Disassembler_dispatch(o);                
            debug_print(">\n");                        
            break;
        }
        case BC_PUSH_BLOCK: {
            debug_print("block: (index: %d) ", BC_1);
            pVMMethod meth = (pVMMethod)SEND(method, get_constant, bc_idx);
            Disassembler_dump_method(meth, "$");
            break;
        }
        case BC_PUSH_CONSTANT: {
            pVMObject constant = SEND(method, get_constant, bc_idx);
            pVMClass c = SEND(constant, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_cname = SEND(cname, get_string);
            debug_print("(index: %d) value: (%s) ", BC_1, 
                        SEND(s_cname, chars));
            _Disassembler_dispatch(constant);
            debug_print("\n");
            break;
        }
        case BC_PUSH_GLOBAL: {
            pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
            pVMObject o = Universe_get_global(name);
            pVMSymbol cname;
            pString s_cname;
            char*   c_cname;
            if(o) {
                pVMClass c = SEND(o, get_class);
                cname = SEND(c, get_name);
                s_cname = SEND(cname, get_string);
                c_cname = (char*)SEND(s_cname, chars);
            } else
                c_cname = "NULL";
            
            pString s_name =SEND(name, get_string);
            debug_print("(index: %d)value: %s <(%s) ", BC_1,
                        SEND(s_name, chars), c_cname);
            _Disassembler_dispatch(o);
            debug_print(">\n");
            break;
        }
        case BC_POP: {
            size_t sp = SEND(frame->stack_pointer, get_embedded_integer);
            pVMObject o = SEND((pVMArray)frame, get_indexable_field, sp);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_cname = SEND(cname, get_string);
            debug_print("popped <(%s) ", SEND(s_cname, chars));
            //dispatch
            _Disassembler_dispatch(o);
            debug_print(">\n");                        
            break;            
        }            
        case BC_POP_LOCAL: {
            size_t sp = SEND(frame->stack_pointer, get_embedded_integer);
            pVMObject o = SEND((pVMArray)frame, get_indexable_field, sp);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_cname = SEND(cname, get_string);
            debug_print("popped local: %d, context: %d <(%s) ", BC_1, BC_2,
                        SEND(s_cname, chars));
            //dispatch
            _Disassembler_dispatch(o);            
            debug_print(">\n");                        
            break;            
        }
        case BC_POP_ARGUMENT: {
            size_t sp = SEND(frame->stack_pointer, get_embedded_integer);
            pVMObject o = SEND((pVMArray)frame, get_indexable_field, sp);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            debug_print("argument: %d, context: %d <(%s) ", BC_1, BC_2,
                        SEND(cname, get_string));
            //dispatch
            _Disassembler_dispatch(o);
            debug_print(">\n");                        
            break;
        }
        case BC_POP_FIELD: {
            size_t sp = SEND(frame->stack_pointer, get_embedded_integer);
            pVMObject o = SEND((pVMArray)frame, get_indexable_field, sp);
            pVMSymbol name = (pVMSymbol)SEND(method, get_constant, bc_idx);
            pVMClass c = SEND(o, get_class);
            pVMSymbol cname = SEND(c, get_name);
            pString s_name = SEND(name, get_string);
            pString s_cname = SEND(cname, get_string);
            debug_print("(index: %d) field: %s <(%s) ",  BC_1,
                        SEND(s_name, chars),
                        SEND(s_cname, chars));
            _Disassembler_dispatch(o);
            debug_print(">\n");                        
            break;
        }
        case BC_SUPER_SEND:
        case BC_SEND: {
            pVMSymbol sel = (pVMSymbol)SEND(method, get_constant, bc_idx);
            pString s_sel = SEND(sel, get_string);

            debug_print("(index: %d) signature: %s (", BC_1,
                        SEND(s_sel, chars));

            //handle primitives, they don't increase call-depth
            pVMFrame frame;
            Interpreter_get_frame(&frame);
            pVMObject elem = SEND(frame, get_stack_element,
                                  Signature_get_number_of_arguments(sel));
            pVMClass elemClass = SEND(elem, get_class);
            pVMObject inv =  SEND(elemClass, lookup_invokable, sel);
            
            if(inv && TSEND(VMInvokable, inv, is_primitive)) 
                debug_print("*)\n");
            else {
                debug_print("\n");    
                indentc++; ikind='>'; // visual
            }
                break;
        }            
        case BC_RETURN_LOCAL:
        case BC_RETURN_NON_LOCAL: {
            debug_print(")\n");
            indentc--; ikind='<'; //visual
            break;
        }
        default:
            debug_print("<incorrect bytecode>\n");
            break;
    }
}

// EOF: diassembler.c

