#ifndef IMAGE_H_
#define IMAGE_H_

/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

/**
 * \file vm/Image.h
 * \brief Header file for the Smalltalk image functionality.
 * 
 * Smalltalk image functionality:
 * A running virtual machine is dumped to a file, the image.
 * At a later time, the image can then be restored into memory and the
 * execution of the VM resumes where it left off.
 * 
 * This file contains the infrastructure to store and restore an image.
 * Necessary data, such as globals and the call stack, is collected by
 * calling the owning components. These then use callbacks.
 *
 * \author Murat Knecht
 * \author Christian Schwarz
 * 
 * \section image_file_format Image file format
 * \verbatim
   ----------------
   | ImageHeader  | = if its possible to read the image at all
   ----------------
   |   Encryption | = optional! if image is encrypted this header 
   |     Header   |   contains the encryption information
   ----------------
   | Compression  | = optional! if image is compressed this header
   |   Header     |   contains the compression information
   ----------------
   | InfoHeader   | = the details to read the actual information of interest 
   ----------------
   | Globals      |
   ----------------
   | Symbol table |
   ----------------
   | Heap         |
   ----------------
   \endverbatim
 */

#include <stdbool.h>
#include <stdlib.h>
#include <time.h>

/// \todo see, if this can be done with a bit less including everything
///
/// 
// #include <vm/Universe.h>

#include <misc/defs.h>

#include <vmobjects/objectformats.h>

/// magic! can be used for endian detection (future versions)
#define IMAGE_MAGIC_CONSTANT     0xCAFE2342
/// Major version is increased, when the image format changes in an
/// incompatible way.
#define IMAGE_VERSION      10


/**
 * ImageEncryptionConstants
 */
#define IMAGE_CRYPT_NONE 12 ///< Image is not encrypted
#define IMAGE_CRYPT_XOR  23 ///< Image is encrypted with XOR
#define IMAGE_CRYPT_RSA  42 ///< Image is encrypted with RSA


/**
 * \typedef ImageHeader
 * \brief First header of image file.
 * 
 * First header of image file which allows the reader to
 * determine whether the image can be loaded.
 */
typedef struct _ImageHeader {
	int32_t magic;          ///< Magic number to determine endianness
	int32_t word_length;    ///< Word length in bit (32 or 64)
	int32_t version; 		///< Version of the image
	int32_t format_info; 	///< information about compression, encryption,... can be saved here
} ImageHeader;

/**
 * \typedef ImageCompressionHeader
 * \brief Contains the requiered compression information
 * \note This header is optional
 */ 
typedef struct _ImageCompressionHeader {
	INTEGER orginal_size;		///< Size of the uncompressed image
	INTEGER compressed_size;    ///< Size of the compressed image
} ImageCompressionHeader;

/**
 * \typedef ImageEncryptionHeader
 * \brief This header contains the requiered information
 * about encryption. Therefore the encryptionmethod is stored
 * \note This header is optional
 */
typedef struct _ImageEncryptionHeader {
	int32_t encryption_method; ///< contains the encryption method
	                           ///  encryption methods are defined in Image.c
							   ///  naming comvention: IMAGE_CRYPT_<METHOD>
} ImageEncryptionHeader;

/**
 * \typedef ImageInfoHeader
 * \brief Info header providing additional data on the image.
 * 
 * More info on the image. Info header may be extended to include
 * user-data, like a description, which makes sense in case of a
 * "minimal snapshot".
 */
typedef struct _ImageInfoHeader {
	time_t timestamp_of_storage;      ///< When the image was stored.
	
	long int vmt_data_offset;         ///< Offset in image file to the data of the universe
	long int gc_data_offset;          ///< Offset in image file to garbage collector data (heap, amongst others)
	long int universe_data_offset;    ///< Offset in image file to the data of the universe
	long int interpreter_data_offset; ///< Offset in image file to interpreter data (currentframe)
} ImageInfoHeader;


/**
 * \brief Store the current VM image in a file.
 * 
 * Store the currently executing VM image in the file with the given name.
 *
 * \param file_name The file name of the file in which to store the image.
 * 		Note that while a relative path may be specified, the working
 * 		directory is chosen on default.
 * 		The ending ".cim" is appended.
 * \param encrypt_alg The algorithm for encrypting the image. (see constants)
 * \param encrypt_key The encryption key, irrelevant in case of IMAGE_CRYPT_NONE;
 * 
 * \return Whether the storage succeeded.
 * 
 * \pre No image storing may be in process.
 */
bool image_store(const pString file_name,
		const int32_t encrypt_alg, const int32_t encrypt_key);

/**
 * \brief Retain info from the image and establish that its loadable.
 * 
 * Inspects the image in the file with the given name. It will be established
 * that the image is loadable by retrieving the ImageHeader and verifying the
 * version.
 * 
 * \param file_name The file name of the image to inspect.
 * 		Note that while a relative path may be specified, the working
 * 		directory is chosen on default.
 * 		The ending ".cim" is appended.
 * 
 * \return The VM object representing the image.
 */
pVMImage image_inspect(const pString image_name);

/**
 * \brief Checks the header, if the image is compressed
 * 
 * The image header is checked, if the format bit for compression
 * is set.
 * 
 * \param image The image under scrunity.
 *
 * \return Whether the image is compressed.
 */
bool image_is_compressed(pVMImage image);

/**
 * \brief Checks the header, if the image is encrypted.
 * 
 * The image header is checked, if the format bit for encryption
 * is set.
 * 
 * \param image The image under scrunity.
 *
 * \return Whether the image is compressed.
 */
bool image_is_encrypted(pVMImage image);

/**
 * \brief Tests a given image, if it is loadable from a VM point of view.
 * 
 * The image is tested for compatibility with the current VM. This includes
 * version checks and, if successful, is an indicator that the image loading
 * will not fail due to the image having the wrong format.
 * 
 * \warning One CANNOT assume successful image loading even if this method
 * returns a positive result. One of arbitrarily many reasons may stand in the way:
 *  * the image format may still be unreadable, if the versioning does not 
 *    accurately reflect the compatibility of the format with the VM.
 *  * mere file access fails, because someone deleted the image
 *  * etc.
 * 
 * \param image The image under scrunity.
 *
 * \return Whether the image is (probably) loadable.
 */
bool image_is_loadable(pVMImage image);

/**
 * \brief Loads a previously inspected image.
 * 
 * Restores a previously inspected image.
 * image_fname
 * 
 * \param image A VM object representing the image to load.
 * \param encrypt_key The encryption key, irrelevant in case of IMAGE_CRYPT_NONE;
 * 
 * \return Whether restoration of the image was successful.
 * 
 * \warning Since the file system usually does not support transactions,
 * neither does the inspect-load sequence: There is no guarantee that a 
 * previously inspected image is loadable. One possibility, out of many,
 * is that the file is deleted after the inspect() but before the load()
 * call. Simpler explanations may include a corrupted file format.
 * This cannot be stressed enough: THE CALLER MUST NOT ASSUME A SUCCESSFUL
 * LOAD CALL!
 */
bool image_load(pVMImage image, const int32_t encrypt_key);

/**
 * \brief Loads the image with the given name.
 * 
 * From the image name, a filename is constructed, and the image file is loaded.
 * 
 * \param image_name The name of the image to load.
 * \param encrypt_key The encryption key, irrelevant in case of IMAGE_CRYPT_NONE;
 * 
 * \return Whether image loading succeeded.
 */
bool image_load_with_name(pString image_name, const int32_t encrypt_key);


/**
 * \brief Loads a sequence of bytes into a possibly permanent buffer.
 * 
 * Loads a number of bytes from the image. The result is stored
 * into a buffer and the number of read bytes is saved in the out parameter.
 * 
 * The buffer is freed when either the function is called the
 * next time or when the serialize function of a component returns. A buffer
 * can be rescued from this fate by demanding that the buffer be made permanent.
 * 
 * \param[out] size The number of bytes
 * \param[in] retain Whether the buffer is to be retained (i.e. made permanent) or not.
 * 
 * \return Pointer to a buffer with size bytes.
 * 
 * \warning When invoked out of image loading context, the function
 * will terminate the VM.
 */
void* image_load_byte_sequence(INTEGER* size, bool retain);

/**
 * \brief Callback function: Stores a sequence of bytes from the caller into the image.
 * 
 * This is a call-back function that may be used by the components
 * involved in the image storage process.
 * Invocation is only allowed as call-back.
 * 
 * \param size Number of bytes to be stored from byte_sequence.
 * \param byte_sequence The data to store.
 * 
 * \warning When invoked out of image persisting context, the function
 * will terminate the VM.
 */
bool image_store_byte_sequence(const INTEGER size, const void* const byte_sequence);

/**
 * \brief Translates the given VMT pointer to a currently valid vmt pointer.
 * 
 * The currently valid pointer to a virtual method table (VMT) is provided. The VMT
 * is identified by the VMT pointer from the previously loaded image.
 * 
 * This function thus provides a mapping from the image VMT pointers to currently valid
 * ones.
 * 
 * \note If no image has been loaded previously, the mapping will reflexive, mapping
 * each object to itself. This allows for testing the relocation of the heap.
 * 
 * \param old The invalid, image-provided VMT pointer.
 * 
 * \return The currently valid VMT pointer.
 */

void* image_translate_vmt_pointer(const void* const old);

#endif /*IMAGE_H_*/
