#ifndef UNIVERSE_H_
#define UNIVERSE_H_

/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <misc/List.h>
#include <misc/debug.h>

#include <vm/Image.h>

#include <vmobjects/OOObject.h>

#include <unistd.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*  Global objects  */
extern pVMObject nil_object;
extern pVMObject true_object;
extern pVMObject false_object;
  
extern pVMClass object_class;
extern pVMClass class_class;
extern pVMClass metaclass_class;
  
extern pVMClass nil_class;
extern pVMClass image_class;
extern pVMClass integer_class;
extern pVMClass biginteger_class;
extern pVMClass array_class;
extern pVMClass method_class;
extern pVMClass symbol_class;
extern pVMClass frame_class;
extern pVMClass primitive_class;
extern pVMClass string_class;
extern pVMClass system_class;
extern pVMClass block_class;
extern pVMClass double_class;

// for runtime debug
extern short dump_bytecodes;
extern short gc_verbosity;
 
 
void          Universe_exit(int);
void          Universe_error_exit(const char* restrict);

void          Universe_assert(bool);

pVMSymbol     Universe_symbol_for(pString);
pVMSymbol     Universe_symbol_for_chars(const char* restrict);

pVMImage      Universe_new_image(pVMInteger, pVMString, ImageHeader, ImageInfoHeader);

void          Universe_initialize_system_class(pVMClass, pVMClass, const char*);

pHashmap      Universe_get_globals_dictionary();
pVMObject     Universe_get_global(pVMSymbol);
void          Universe_set_global(pVMSymbol, pVMObject);
bool          Universe_has_global(pVMSymbol);

pVMClass      Universe_get_block_class(void);
pVMClass      Universe_get_block_class_with_args(size_t);

pVMClass      Universe_load_class(pVMSymbol);
void          Universe_load_system_class(pVMClass);
pVMClass      Universe_load_class_basic(pVMSymbol, pVMClass);
pVMClass      Universe_load_shell_class(pString);

pString*      Universe_handle_arguments(int* vm_argc, int argc, char** argv) ;
void          Universe_initialize(int argc, pString* argv);
void          Universe_destruct(void);

bool          Universe_is_global(const pVMObject subject);
bool          Universe_is_system_class(const pVMClass subject);

/**
 * \brief Update references of globals and symbols
 * 
 * Update references to GC-managed objects in case of a
 * relocated or loaded heap.
 * 
 * \param difference The difference by which to change the pointers.
 */ 
void          Universe_update_references(const INTEGER difference);

/**
 * \brief Save the Universe's state in the image file.
 * 
 * Save the state of Universe in the image file.
 * 
 * \return True in case of successful execution, false otherwise.
 */
bool          Universe_serialize(void);

/**
 * \brief Load Universe's state from the image file.
 * 
 * Restore a previously stored state from an image file.
 * Take the difference in heap-addresses into account
 * and get globals and symbols ready.
 * 
 * \param difference The difference by which to update the object references.
 * \return True in case of successful execution, false otherwise.
 */
bool          Universe_deserialize(INTEGER difference);

/**
 * \brief Store object-owned non-heap data of globals and symbols.
 * 
 * Make sure that globals and symbols, being heap objects, store their
 * non-heap data in the image file.
 * 
 * \note May be removed when objects are fully heap-contained.
 * 
 * \return True in case of successful execution, false otherwise.
 */
bool 		  Universe_serialize_object_data(void);

/**
 * \brief Load object-owned non-heap data for globals and symbols.
 * 
 * Make sure that globals and symbols, being heap objects, retrieve their
 * non-heap data from the image file.
 * 
 * \note May be removed when objects are fully heap-contained.
 * 
 * \return True in case of successful execution, false otherwise.
 */
bool          Universe_deserialize_object_data(void);

/**
 * \brief Register a given class as having loaded primitive code.
 * 
 * Calling this method registers the given class as having dynamically
 * loaded code. This is important to know for image functionality, as
 * after an image has been loaded and its contents restored, any previously
 * loaded and now assumed-to-be-available code must again be loaded from
 * the libraries.
 * 
 * \param class The class to register as having primitives.
 */
void Universe_register_class_with_primitives(const pVMClass);

/**
 * \brief Stores references to classes with primitives.
 * 
 * This class stores class references previously registered as having
 * primitives in the image file.
 * 
 * \return True in case of successful execution, false otherwise.
 * 
 * \see Universe_register_class_with_primitives
 */
bool Universe_store_primitives(void);

/**
 * \brief Makes sure classes with primitives load their code after image loading.
 * 
 * Restores the list of classes having primitives. It then makes sure
 * that these classes load their code (for primitives) from the libraries.
 * 
 * \return True in case of successful execution, false otherwise.
 * 
 * \see Universe_register_class_with_primitives
 */
bool Universe_finalize(INTEGER difference);

#endif UNIVERSE_H_
