/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */
 
#include "Symboltable.h"
#include "VMObject.h"
#include "VMSymbol.h"
#include "VMInvokable.h"
#include "VMClass.h"

#include <memory/memory.h>

#include <vm/Universe.h>

#include <interpreter/Interpreter.h>

#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>

/// some constants (as actual constants, so they end up in the compiler's symbol table, which eases debugging)
const int32_t REFERENCES_UPDATED = -10;
const int32_t OBJECT_DATA_SERIALIZED = -20;
const int32_t OBJECT_DATA_RESTORED = -30;


// number of fields helper
void set_number_of_fields(void* _self, int value);


//
//  Class Methods (Starting with VMObject_) 
//

/**
 * Create a new VMObject
 */

pVMObject VMObject_new(void) {
    return VMObject_new_num_fields(NUMBER_OF_OBJECT_FIELDS(VMObject));
}

/**
 * Create a new VMObject with a specific number of fields
 */
pVMObject VMObject_new_num_fields(size_t number_of_fields) {
    size_t object_stub_size = 
        sizeof(VMObject) - 
        sizeof(pVMObject) * NUMBER_OF_OBJECT_FIELDS(VMObject);
    pVMObject result = (pVMObject)memory_allocate_object(
        object_stub_size + sizeof(pVMObject) * number_of_fields);
    if(result) {
        result->_vtable = VMObject_vtable();
        INIT1(result, &number_of_fields);
    }
    return result;
}


pVMObject VMObject_new_instance(pVMClass instance_class) {
    // Allocate a new instance and set its class to be the given class
    pVMObject result = VMObject_new_num_fields(
        SEND(instance_class, get_number_of_instance_fields));
    SEND(result, set_class, instance_class);
    
    // Return the freshly allocated instance
    return result;
}


/**
 * Initialize a VMObject
 */
void _VMObject_init(void* _self, void* field_count, void* param2) {
    pVMObject self = (pVMObject)_self;
    SUPER(OOObject, self, init, NULL, NULL);
    
    set_number_of_fields(self, *(size_t*)field_count);
}


void _VMObject_free(void* _self) {
    memory_free(_self);
}


//
//  Instance Methods (Starting with _VMObject_) 
//


pVMObject _VMObject_get_field(void* _self, int index) {
    pVMObject self = (pVMObject)_self;
    // get the field with the given index
    return self->fields[index];
}


void _VMObject_set_field(void* _self, int index, pVMObject value) {
    pVMObject self = (pVMObject)_self;
    // set the field with the given index to the given value
    self->fields[index] = value;
}


pVMClass _VMObject_get_class(void* _self) {
    pVMObject self = (pVMObject)_self;
    // get the class of this object
    return self->clazz;
}


void _VMObject_set_class(void* _self, pVMClass value) {
    pVMObject self = (pVMObject)_self;
    // set class.
    // this is equivalent to setting the Class Index Field
    self->clazz = value;
}


pVMSymbol _VMObject_get_field_name(void* _self, int index) {
    pVMObject self = (pVMObject)_self;
    // get the name of the field with the given index
    return SEND(SEND(self,get_class), get_instance_field_name, index);
}


int _VMObject_get_field_index(void* _self, pVMSymbol name) {
    pVMObject self = (pVMObject)_self;
    // get the index for the field with the given name
    return SEND(SEND(self,get_class), lookup_field_index, name);
}


int _VMObject_get_number_of_fields(void* _self) {
    pVMObject self = (pVMObject)_self;
    // get the number of fields in this object
    return self->num_of_fields;
}


// number of fields helper
void set_number_of_fields(void* _self, int value) {
    pVMObject self = (pVMObject)_self;
    self->num_of_fields = value;
    
    // clear each and every field by putting nil into them
    for(int i = 0; i < value; i++) {
        SEND(self, set_field, i, nil_object);
    }
}


void _VMObject_send(void* _self, pString selector_string, 
                    pVMObject* arguments, int argc) {
    pVMObject self = (pVMObject)_self;
    // turn the selector string into a selector
    pVMSymbol selector = Universe_symbol_for(selector_string);
    // push the receiver onto the stack
    pVMFrame frame;
    Interpreter_get_frame(&frame);
    SEND(frame, push, self);
    
    // push the arguments onto the stack
    for(int i = 0; i < argc; i++) {
          SEND(frame, push, arguments[i]);
    }
    // lookup the invokable
    pVMClass clazz = SEND(self, get_class);
    pVMObject invokable = SEND(clazz, lookup_invokable, selector);
    // invoke the invokable
    TSEND(VMInvokable, invokable, invoke, frame);
}


void VMObject_assert(bool value) {
    Universe_assert(value);
}


void _VMObject_mark_references(void* _self) {
    pVMObject self = (pVMObject) _self;
    gc_mark_object(self->clazz);
    for(int i = 0;i < SEND(self, get_number_of_fields); i++) {
        gc_mark_object(SEND(self, get_field, i));
    }
}

// enforces checks previous to object processing
void update_object_references(pVMObject obj, const INTEGER difference) {
	
	// If its not in the heap, its not a managed object and MUST NOT
	// be relocated.
	if (!gc_is_managed_object(obj))
		return;
	
	// Return if the object has already been processed.
	if (REFERENCES_UPDATED == obj->gc_field)
		return;
	
	// update vtable pointer and continue traversing the object tree, updating heap-pointer
	obj->_vtable = (VTABLE(VMObject)*)image_translate_vmt_pointer(obj->_vtable);
	SEND(obj, update_references, difference);
}

/**
 * \brief Updates all pointers to referenced heap-located objects.
 * 
 * In case of a loaded or relocated heap, all object references in 
 * this particular object are updated. The method is invoked recursively
 * on all these objects.
 * 
 * \param _self The object in question.
 * \param difference The difference by which to change the object pointers.
 */
void _VMObject_update_references(void* _self, const INTEGER difference) {
	pVMObject self = (pVMObject) _self;

	// mark object as processed
	self->gc_field = REFERENCES_UPDATED;
	
	// iterate through fields and update them
	int number = SEND(self, get_number_of_fields);
	for (int curr_field_idx = 0; curr_field_idx < number; curr_field_idx++) {
		
		pVMObject old = SEND(self, get_field, curr_field_idx);
		pVMObject anew = (pVMObject)((INTEGER)old + difference);
		
		// managed object?
		if (!gc_is_managed_object(anew))
			continue;
		
		SEND(self, set_field, curr_field_idx, anew);
		update_object_references(anew, difference);
	}
}

// enforces checks previous to object processing
bool serialize_object_data(pVMObject obj) {
	
	// If its not in the heap, its not a managed object and MUST NOT
	// be relocated.
	if (!gc_is_managed_object(obj))
		return true;
	
	// Return if the object has already been processed.
	if (OBJECT_DATA_SERIALIZED == obj->gc_field)
		return true;
	
	// update vtable pointer and continue traversing the object tree, updating heap-pointer
	return SEND(obj, serialize_data);
}


/**
 * \brief Store the object's non-heap data to the image file.
 * 
 * The object saves its non-heap data, i.e. strings in the image file.
 * The method is invoked recursively for all referenced objects.
 * 
 * \param _self The object in question.
 * \return True in case of successful saving of data, false otherwise.
 * 
 * \note This method may be removed when all objects are fully heap-contained.
 * \note This method is meant to be overridden, as an object of VMObject
 * does not have any data to be serialized.
 */
bool _VMObject_serialize_data(void* _self) {
	pVMObject self = (pVMObject)_self;
	
	self->gc_field = OBJECT_DATA_SERIALIZED;
	
	// iterate through fields and update them
	// skip class field
	int number = SEND(self, get_number_of_fields);
	for (int curr_field_idx = 0; curr_field_idx < number; ++curr_field_idx) {

		if (!serialize_object_data(SEND(self, get_field, curr_field_idx)))
				return false;
	}

	return true;
}

// enforces checks previous to object processing
bool deserialize_object_data(pVMObject obj) {
	
	// If its not in the heap, its not a managed object and MUST NOT
	// be relocated.
	if (!gc_is_managed_object(obj))
		return true;
	
	// Return if the object has already been processed.
	if (OBJECT_DATA_RESTORED == obj->gc_field)
		return true;
	
	// update vtable pointer and continue traversing the object tree, updating heap-pointer
	return SEND(obj, deserialize_data);
}


/**
 * \brief Deserializes the object's non-heap data from the image file.
 * 
 * The object loads its non-heap data, i.e. strings from the image file.
 * The method is invoked recursively for all referenced objects.
 * 
 * \param _self The object in question.
 * \return True in case of successful restoration of data, false otherwise.
 * 
 * \note This method may be removed when all objects are fully heap-contained.
 * \note This method is meant to be overridden, as an object of VMObject
 * does not have any data to be serialized.
 */
bool _VMObject_deserialize_data(void* _self) {
	pVMObject self = (pVMObject)_self;
	
	self->gc_field = OBJECT_DATA_RESTORED;

	// iterate through fields and update them
	// skip class field
	int number = SEND(self, get_number_of_fields);
	for (int curr_field_idx = 0; curr_field_idx < number; ++curr_field_idx) {

		if (!deserialize_object_data(SEND(self, get_field, curr_field_idx)))
			return false;
	}

	return true;
}


//
// The VTABLE-function
//


static VTABLE(VMObject) _VMObject_vtable;
bool VMObject_vtable_inited = false;


VTABLE(VMObject)* VMObject_vtable(void) {
    if(! VMObject_vtable_inited) {
        *((VTABLE(OOObject)*)&_VMObject_vtable) = *OOObject_vtable();
        _VMObject_vtable.init = METHOD(VMObject, init);        
        _VMObject_vtable.free = METHOD(VMObject, free);        
        _VMObject_vtable.get_class = METHOD(VMObject, get_class);
        _VMObject_vtable.set_class = METHOD(VMObject, set_class);
        _VMObject_vtable.get_field_name = METHOD(VMObject, get_field_name);
        _VMObject_vtable.get_field_index = METHOD(VMObject, get_field_index);
        _VMObject_vtable.get_number_of_fields =
            METHOD(VMObject, get_number_of_fields);
        _VMObject_vtable.send = METHOD(VMObject, send);
        _VMObject_vtable.get_field = METHOD(VMObject, get_field);
        _VMObject_vtable.set_field = METHOD(VMObject, set_field);
        _VMObject_vtable.update_references = METHOD(VMObject, update_references);
        _VMObject_vtable.serialize_data = METHOD(VMObject, serialize_data);
        _VMObject_vtable.deserialize_data = METHOD(VMObject, deserialize_data);
        
        VMObject_vtable_inited = true;
    }
    return &_VMObject_vtable;
}
