/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include "VMPrimitive.h"
#include "VMInvokable.h"
#include "VMSymbol.h"

#include <memory/memory.h>

#include <misc/debug.h>

#include <vm/Universe.h>

#include <stdbool.h>


void empty_routine(pVMObject prim, pVMFrame);


//
//  Class Methods (Starting with VMPrimitive_) 
//


/**
 * Create a new VMPrimitive
 */
pVMPrimitive VMPrimitive_new(pVMSymbol signature) {
    pVMPrimitive result = (pVMPrimitive)memory_allocate_object(sizeof(VMPrimitive));
    if(result) {
        result->_vtable = VMPrimitive_vtable();
        size_t size = NUMBER_OF_OBJECT_FIELDS(VMPrimitive);
        INIT2(result, &size, signature);
    }
    
    return result;
}


/**
 * Create a new VMPrimitive with an explicit empty-routine
 */
pVMPrimitive VMPrimitive_get_empty_primitive(pVMSymbol signature) {
    debug_info("\t\t'%s'*\n", signature->embedded_string->chars);
    pVMPrimitive prim = VMPrimitive_new(signature);
    *(prim->empty) = true;
    prim->routine = empty_routine;
    
    // Return an empty primitive with the given signature
    return prim;
}


pVMPrimitive VMPrimitive_assemble(method_generation_context* mgenc) {
    return VMPrimitive_get_empty_primitive(mgenc->signature);
}


/** 
 * Initialize a VMPrimitive
 */
void _VMPrimitive_init(void* _self, void* param1, void* param2) {
    pVMPrimitive self = (pVMPrimitive)_self;
   
    // call "super" constructor
    SUPER(VMObject, self, init, param1, NULL);
    
    //
    // Set the class of this primitive to be the universal primitive class
    // NOTE: this is the only branch of the object hierachy, which 
    //       does this in the init method!
    //
    SEND(self, set_class, primitive_class);
    
    // Set the signature of this primitive
    TSEND(VMInvokable, self, set_signature, (pVMSymbol)param2);
    
    // set routine to NULL
    self->routine = NULL;
    self->empty = (bool*)memory_allocate(sizeof(bool) + PAD_BYTES(sizeof(bool)));
    *(self->empty) = false;
}


void _VMPrimitive_free(void* _self) {
    pVMPrimitive self = (pVMPrimitive)_self;
    if(self->empty)
        memory_free(self->empty);
    SUPER(VMObject, self, free);
}


//
//  Instance Primitive (Starting with _VMPrimitive_) 
//


bool _VMPrimitive_is_empty(void* _self) {
    pVMPrimitive self = (pVMPrimitive)_self;
    return *(self->empty);
}


void _VMPrimitive_set_routine(void* _self, routine_fn routine) {
    pVMPrimitive self = (pVMPrimitive)_self;
    self->routine = routine;
}


void empty_routine(pVMObject self, pVMFrame frame) {
    pVMSymbol sig = TSEND(VMInvokable, self, get_signature);
    pString str = SEND(sig, get_string);
    debug_warn("undefined primitive %s called", SEND(str, chars));
}



void _VMPrimitive_update_references(void* _self, const INTEGER difference) {
	pVMPrimitive self = (pVMPrimitive)_self;
	
	int nof = self->num_of_fields;
	self->num_of_fields = 3;
	SUPER(VMObject, self, update_references, difference);
	self->num_of_fields = nof;
}

bool _VMPrimitive_serialize_data(void* _self) {
	pVMPrimitive self = (pVMPrimitive)_self;

	int nof = self->num_of_fields;
	self->num_of_fields = 3;
	if (!SUPER(VMObject, self, serialize_data))
		return false;
	self->num_of_fields = nof;
	
	return true;
}

bool _VMPrimitive_deserialize_data(void* _self) {
	pVMPrimitive self = (pVMPrimitive)_self;
	
	int nof = self->num_of_fields;
	self->num_of_fields = 3;
	if (!SUPER(VMObject, self, deserialize_data))
		return false;
	self->num_of_fields = nof;
	
	return true;
}


//
// The VTABLE-function
//


static VTABLE(VMPrimitive) _VMPrimitive_vtable;
bool VMPrimitive_vtable_inited = false;


VTABLE(VMPrimitive)* VMPrimitive_vtable(void) {
    if(! VMPrimitive_vtable_inited) {
        *((VTABLE(VMObject)*)&_VMPrimitive_vtable) = *VMObject_vtable();

        ASSIGN_TRAIT(VMInvokable, VMPrimitive);

        _VMPrimitive_vtable.init        = METHOD(VMPrimitive, init);
        _VMPrimitive_vtable.is_empty    = METHOD(VMPrimitive, is_empty);
        _VMPrimitive_vtable.set_routine = METHOD(VMPrimitive, set_routine);
        _VMPrimitive_vtable.free        = METHOD(VMPrimitive, free);

        _VMPrimitive_vtable.update_references = METHOD(VMPrimitive, update_references);
        _VMPrimitive_vtable.serialize_data    = METHOD(VMPrimitive, serialize_data);
        _VMPrimitive_vtable.deserialize_data  = METHOD(VMPrimitive, deserialize_data);

        VMPrimitive_vtable_inited = true;
    }
    return &_VMPrimitive_vtable;
}
