/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */
 
#include "VMClass.h"
#include "VMInvokable.h"
#include "VMMethod.h"
#include "VMSymbol.h"
#include "VMArray.h"
#include "VMPrimitive.h"

#include <memory/memory.h>

#include <gc-refcount/gc.h>

#include <misc/debug.h>
#include <misc/defs.h>
#include <misc/String.h>

#include <vm/Universe.h>

#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>

#ifdef CSOM_WIN
/**
 * Emualting the dl-interface with win32 means
 */
#   define WIN32_LEAN_AND_MEAN
#   define dlerror()   "Load Error"
#   define dlsym       GetProcAddress
#   define DL_LOADMODE NULL, LOAD_WITH_ALTERED_SEARCH_PATH
#   define dlopen      LoadLibraryEx
#   define dlclose     CloseHandle
#   include <windows.h>
#else
#   include <dlfcn.h>
#endif

// private
int number_of_super_instance_fields(void* _self);

//
//  Class Methods (Starting with VMClass_) 
//


/**
 * Create a new VMClass
 */
pVMClass VMClass_new(void) {
    return VMClass_new_num_fields(NUMBER_OF_OBJECT_FIELDS(VMClass));
}


pVMClass VMClass_new_with(pVMClass class_of_class) {
    // Allocate a new class and set its class to be the given class class
    int num_fields = SEND(class_of_class, get_number_of_instance_fields);
    pVMClass result;
    
    if(num_fields) //this is a normal class as class class
        result = VMClass_new_num_fields(num_fields);        
    else // ths class_of_class has no fields
        result = VMClass_new();

    SEND((pVMObject)result, set_class, class_of_class);
    
    // Return the freshly allocated class
    return result;
}


/**
 * Create a new VMClass with a specific number of fields
 */
pVMClass VMClass_new_num_fields(size_t number_of_fields) {
    // calculate Class size without fields
    size_t class_stub_size = sizeof(VMClass) - 
                             sizeof(pVMObject) * (NUMBER_OF_OBJECT_FIELDS(VMClass));
    pVMClass result = 
        (pVMClass)memory_allocate_object(class_stub_size +
                             sizeof(pVMObject) * number_of_fields);

    if(result) {
        result->_vtable = VMClass_vtable();
        INIT(result, number_of_fields);
    }

    return result;
}


pVMClass VMClass_assemble(class_generation_context* cgc) {
    // build class class name
    pString ccname = String_new_from(SEND(cgc->name, get_string));
    SEND(ccname, concatChars, " class");
    
    // Load the super class
    pVMClass super_class = Universe_load_class(cgc->super_name);
    
    // Allocate the class of the resulting class
    pVMClass result_class = VMClass_new_with(metaclass_class);

    // Initialize the class of the resulting class
    SEND(result_class, set_instance_fields,
    		VMArray_new_from_list(cgc->class_fields));
    SEND(result_class, set_instance_invokables,
    		VMArray_new_from_list(cgc->class_methods));
    SEND(result_class, set_name, Universe_symbol_for(ccname));
    pVMClass super_mclass = SEND(super_class, get_class);
    SEND(result_class, set_super_class, super_mclass);
    
    // Allocate the resulting class
    pVMClass result = VMClass_new_with(result_class);
    
    // Initialize the resulting class
    SEND(result, set_instance_fields,
    	VMArray_new_from_list(cgc->instance_fields));
    SEND(result, set_instance_invokables,
    	VMArray_new_from_list(cgc->instance_methods));
    SEND(result, set_name, cgc->name);
    SEND(result, set_super_class, super_class);
    
    return result;
}


void VMClass_assemble_system_class(class_generation_context* cgc,
                                   pVMClass system_class
) {
    //System classes have only to be filed with fields and methods.
        
    // instance-bound
    SEND(system_class, set_instance_invokables, 
    	VMArray_new_from_list(cgc->instance_methods));
    SEND(system_class, set_instance_fields,
    	VMArray_new_from_list(cgc->instance_fields));
    // class-bound == class-instance-bound 
    pVMClass super_mclass = SEND(system_class, get_class);
    SEND(super_mclass, set_instance_invokables,
         VMArray_new_from_list(cgc->class_methods));
    SEND(super_mclass, set_instance_fields,
         VMArray_new_from_list(cgc->class_fields));
}


//
//  Instance Methods (Starting with _VMClass_) 
//
pVMClass _VMClass_get_super_class(void* _self) {
    pVMClass self = (pVMClass)_self;
    // get the super class
    return self->super_class;
}


void _VMClass_set_super_class(void* _self, pVMClass value) {
    pVMClass self = (pVMClass)_self;
    // set the super class

    gc_inc_reference_counter((pVMObject)value);
    gc_dec_reference_counter((pVMObject)self->super_class);
    self->super_class = value;
}


bool _VMClass_has_super_class(void* _self) {
    pVMClass self = (pVMClass)_self;
    // check whether or not this class has a super class
    return (pVMObject)self->super_class != nil_object;
}


pVMSymbol _VMClass_get_name(void* _self) {
    pVMClass self = (pVMClass)_self;
    // get the name of this class
    return self->name;
}


void _VMClass_set_name(void* _self, pVMSymbol value) {
    pVMClass self = (pVMClass)_self;
    // set the name of this class by writing to the field with name index
    gc_inc_reference_counter((pVMObject)value);
    gc_dec_reference_counter((pVMObject)self->name);
    self->name = value;
}


pVMArray _VMClass_get_instance_fields(void* _self) {
    pVMClass self = (pVMClass)_self;
    // get the instance fields
    return self->instance_fields;
}


void _VMClass_set_instance_fields(void* _self, pVMArray value) {
    pVMClass self = (pVMClass)_self;
    // set the instance fields
    gc_inc_reference_counter((pVMObject)value);
    gc_dec_reference_counter((pVMObject)self->instance_fields);
    self->instance_fields = value;
}


pVMArray _VMClass_get_instance_invokables(void* _self) {
    pVMClass self = (pVMClass)_self;
    // get the instance invokables
    return self->instance_invokables;
}


void _VMClass_set_instance_invokables(void* _self, pVMArray value) {
    pVMClass self = (pVMClass)_self;
    // set the instance invokables 
    gc_inc_reference_counter((pVMObject)value);

    pVMObject oldValue = (pVMObject)self->instance_invokables;

    self->instance_invokables = value;
    
    // make sure this class is the holder of all invokables in the array
    for(int i = 0; i < SEND(self, get_number_of_instance_invokables); i++) {
        pVMObject inv = SEND(self, get_instance_invokable, i);
        TSEND(VMInvokable, inv, set_holder, self);
    }

    gc_dec_reference_counter(oldValue);
}


int _VMClass_get_number_of_instance_invokables(void* _self) {
    pVMClass self = (pVMClass)_self;
    // return the number of instance invokables in this class
    pVMArray arr = SEND(self, get_instance_invokables);
    return SEND(arr, get_number_of_indexable_fields);
}


pVMObject _VMClass_get_instance_invokable(void* _self, int index) {
    pVMClass self = (pVMClass)_self;
    // get the instance invokable with the given index
    pVMArray arr = SEND(self, get_instance_invokables);
    return SEND(arr, get_indexable_field, index);
}


void _VMClass_set_instance_invokable(void* _self, int idx, pVMObject value) {
    pVMClass self = (pVMClass)_self;
    // set this class as the holder of the given invokable
    TSEND(VMInvokable, value,  set_holder, self);
    // set the instance method with the given index to the given value
    pVMArray arr = SEND(self, get_instance_invokables);
    SEND(arr, set_indexable_field, idx, value);
}


pVMObject _VMClass_lookup_invokable(void* _self, pVMSymbol signature) {
    pVMClass self = (pVMClass)_self;
    pVMObject invokable=NULL;
    // lookup invokable with given signature in array of instance invokables
    for(int i = 0; i < SEND(self, get_number_of_instance_invokables); i++) {
      // get the next invokable in the instance invokable array
      invokable = SEND(self, get_instance_invokable, i);
      // return the invokable if the signature matches
      if(TSEND(VMInvokable, invokable, get_signature) == signature)
          return invokable;
    }    
    // traverse the super class chain by calling lookup on the super class
    if(SEND(self, has_super_class)) {
        invokable = SEND(self->super_class, lookup_invokable, signature);
        if(invokable)
            return invokable;
    }
    // invokable not found
    return NULL;
}


int _VMClass_lookup_field_index(void* _self, pVMSymbol field_name) {
    pVMClass self = (pVMClass)_self;
    // lookup field with given name in array of instance fields
    for(int i = SEND(self, get_number_of_instance_fields) - 1; i >= 0; i--) {
      // return the current index if the name matches
      if(field_name == SEND(self, get_instance_field_name, i))
        return i;
    }
    // field not found
    return -1;
}


bool _VMClass_add_instance_invokable(void* _self, pVMObject value) {
    pVMClass self = (pVMClass)_self;
    // add the given invokable to the array of instance invokables
    for(int i = 0; i < SEND(self, get_number_of_instance_invokables); i++) {
        // get the next invokable in the instance invokable array
        pVMObject invokable = SEND(self, get_instance_invokable, i);
        // replace the invokable with the given one if the signature matches
        if(TSEND(VMInvokable, invokable, get_signature) 
           == TSEND(VMInvokable, value, get_signature)) {
            SEND(self, set_instance_invokable, i, value);
            return false;
        }
    }
    // append the given method to the array of instance methods
    self->instance_invokables = 
        SEND(self->instance_invokables, copy_and_extend_with, value);
    return true;
}


void _VMClass_add_instance_primitive(void* _self, pVMPrimitive value) {
    pVMClass self = (pVMClass)_self;
    if(SEND(self, add_instance_invokable, (pVMObject)value)) {
        pVMSymbol sym = TSEND(VMInvokable, value, get_signature);
        debug_warn("Primitive %s is not in class definition for class %s.\n",
                   sym->embedded_string->chars,
                   self->name->embedded_string->chars);
    }
}


pVMSymbol _VMClass_get_instance_field_name(void* _self, int index) {
    pVMClass self = (pVMClass)_self;
    // get the name of the instance field with the given index
    if(index >= number_of_super_instance_fields(self)) {
        // adjust the index to account for fields defined in the super class
        index -= number_of_super_instance_fields(self);
        // return the symbol representing the instance fields name
        return 
            (pVMSymbol)SEND(self->instance_fields, get_indexable_field, index);
    } else {
        // ask the super class to return the name of the instance field
        return SEND(self->super_class, get_instance_field_name, index);
    }
}


int _VMClass_get_number_of_instance_fields(void* _self) {
    pVMClass self = (pVMClass)_self;
    // get the total number of instance fields in this class
    return
        SEND(self->instance_fields, get_number_of_indexable_fields) +
        number_of_super_instance_fields(self);
}


bool _VMClass_has_primitives(void* _self) {
    pVMClass self = (pVMClass)_self;
    // lookup invokable with given signature in array of instance invokables
    for(int i = 0; i < SEND(self, get_number_of_instance_invokables); i++) {
      // get the next invokable in the instance invokable array
        pVMObject inv = SEND(self, get_instance_invokable, i);
        if(TSEND(VMInvokable, inv, is_primitive))
            return true;
    }
    return false;  
}


int number_of_super_instance_fields(void* _self) {
    pVMClass self = (pVMClass)_self;
    /*
     * get the total number of instance fields defined in super classes
     */
    if(SEND(self, has_super_class))
        return SEND(self->super_class, get_number_of_instance_fields);
    else
        return 0;
}


#pragma mark primitive loading helpers


/*
 * these functions are used solely in VMClass: load_primitives. 
 */

/**
 * generate the string containing the path to a primitive library which may be 
 * located at the classpath given.
 */
pString gen_loadstring(const pString restrict cp, 
                       const pString restrict cname
                       ) {
   
    pString loadstring = String_new_from(cp);
    SEND(loadstring, concatChars, file_separator);
    SEND(loadstring, concat, cname);
    SEND(loadstring, concatChars, shared_extension);
    
    return loadstring;
}


/**
 *  generate the string containing the path to a SOMCore which may be located
 *  at the classpath given.
 *
 */
pString gen_core_loadstring(const pString restrict cp) {
    #define S_CORE "SOMCore"
    pString corename = String_new(S_CORE);
    pString result = gen_loadstring(cp, corename);
    SEND(corename, free);
    
    return result;
}


/**
 * load the given library, return the handle
 *
 */
void* load_lib(const pString restrict path) {

    #if !defined(CSOM_WIN)
        #ifdef DEBUG
            #define    DL_LOADMODE RTLD_NOW
        #else
            #define    DL_LOADMODE RTLD_LAZY
        #endif DEBUG
    #endif
    
    // static handle. will be returned
    static void* handle = NULL;
    
	// try load lib
	if((handle=dlopen(SEND(path, chars), DL_LOADMODE)))
		//found.
		return handle;
	else
        return NULL;
}


/**
 * check, whether the lib referenced by the handle supports the class given
 */
bool is_responsible(void* handle, const pString restrict clazz) {
    // function handler
    supports_class_fn supports_class=NULL;

    supports_class = (supports_class_fn)dlsym(handle, "supports_class");

	if(!supports_class) {
        debug_error(dlerror());
        Universe_error_exit("Library doesn't have expected format");
    }
    
    // test class responsibility
    return (*supports_class)(SEND(clazz, chars));
}


/*
 * Format definitions for Primitive naming scheme.
 */
#define CLASS_METHOD_FORMAT_S "%s_%s"
// as in AClass_aClassMethod
#define INSTANCE_METHOD_FORMAT_S "_%s_%s"
// as in _AClass_anInstanceMethod


/*
 * set the routines for primitive marked invokables of the given class
 */
void set_primitives(pVMClass clazz, void* handle, const pString restrict cname,
                    const char* restrict format
                    ) {    
    pVMPrimitive the_primitive;
    routine_fn   routine=NULL;
    
    // iterate invokables
    for(int i = 0; i < SEND(clazz, get_number_of_instance_invokables); i++) {
        the_primitive = (pVMPrimitive)SEND(clazz, get_instance_invokable, i);
        
        if(TSEND(VMInvokable, the_primitive, is_primitive)) {
            // we have a primitive to load
            // get it's selector
            pVMSymbol sig = TSEND(VMInvokable, the_primitive, get_signature);
            pString selector = SEND(sig, get_plain_string);
            
			{ //string block                
				char symbol[SEND(cname, length) + SEND(selector, length)+2 + 1];
                                                                //2 for 2x '_'
				sprintf(symbol, format,
					SEND(cname, chars),
					SEND(selector, chars));
                 
				// try loading the primitive
				routine = (routine_fn)dlsym(handle, symbol);
            }
            
            if(!routine) {
                debug_error("could not load primitive '%s' for class %s\n"
                            "ERR: routine not in library\n",
                            SEND(selector, chars),
                            SEND(cname, chars));
                Universe_exit(ERR_FAIL);
            }
            
            // set routine
            SEND(the_primitive, set_routine, routine);
            the_primitive->empty = false;
        }
    }
}


/**
 * Load all primitives for the class given _and_ its metaclass
 */
void _VMClass_load_primitives(void* _self, const pString* cp, int cp_count) {
    pVMClass self = (pVMClass)_self;
    // the library handle
    void* dlhandle = NULL;
    // cached object properties
    pString cname = SEND(self->name, get_string);

    // iterate the classpathes
    for(int i = 0; (i < cp_count) && !dlhandle; i++) {
        // check the core library
        pString loadstring = gen_core_loadstring(cp[i]);
        dlhandle = load_lib(loadstring);
        SEND(loadstring, free);
        if(dlhandle && is_responsible(dlhandle, cname))
            // the core library is found and responsible
            break;
        
        // the core library is not found or respondible, 
        // continue w/ class file
        loadstring = gen_loadstring(cp[i], cname);
        dlhandle = load_lib(loadstring);
        SEND(loadstring, free);
        if(dlhandle) {
            // the class library was found...
            if(is_responsible(dlhandle, cname)) {
                // ...and is responsible:
                // explicitely call the libirary initializer
            	void* init_csp = dlsym(dlhandle, "init_csp");
            	if (init_csp) {
            		((init_csp_fn)init_csp)();
            	} else {
            		dlclose(dlhandle);
            		Universe_error_exit("Library does not define the init_csp() initializer.");
            	}
                break;
            } else {
                // ... but says not responsible, but have to
                // close it
                dlclose(dlhandle);
                Universe_error_exit("Library claims no responsiblity, "
                                    "but musn't!");        
            }
        }// continue checking the next class path
    }

    // finished cycling,
    // check if a lib was found.
    if(!dlhandle) {
        debug_error("load failure: %s\n", dlerror());
        debug_error("could not load primitive library for %s\n",
                    SEND(cname, chars));
        Universe_exit(ERR_FAIL);
    }
    // do the actual loading for both class and metaclass
    set_primitives(self, dlhandle, cname, INSTANCE_METHOD_FORMAT_S);
    set_primitives(self->clazz, dlhandle, cname, CLASS_METHOD_FORMAT_S);
}


//
// The VTABLE-function
//
//
static VTABLE(VMClass) _VMClass_vtable;
bool VMClass_vtable_inited = false;

VTABLE(VMClass)* VMClass_vtable(void) {
    if(VMClass_vtable_inited) 
        return &_VMClass_vtable;

    *((VTABLE(VMObject)*)&_VMClass_vtable) = *VMObject_vtable();
    _VMClass_vtable.get_super_class = METHOD(VMClass, get_super_class);
    _VMClass_vtable.set_super_class = METHOD(VMClass, set_super_class);
    _VMClass_vtable.has_super_class = METHOD(VMClass, has_super_class);
    _VMClass_vtable.get_name        = METHOD(VMClass, get_name);
    _VMClass_vtable.set_name        = METHOD(VMClass, set_name);
    _VMClass_vtable.get_instance_fields = METHOD(VMClass, get_instance_fields);
    _VMClass_vtable.set_instance_fields = METHOD(VMClass, set_instance_fields);
    _VMClass_vtable.get_instance_invokables = 
        METHOD(VMClass, get_instance_invokables);
    _VMClass_vtable.set_instance_invokables = 
        METHOD(VMClass, set_instance_invokables);
    _VMClass_vtable.get_number_of_instance_invokables =
        METHOD(VMClass, get_number_of_instance_invokables);
    _VMClass_vtable.get_instance_invokable = 
        METHOD(VMClass, get_instance_invokable);
    _VMClass_vtable.set_instance_invokable =
        METHOD(VMClass, set_instance_invokable);
    _VMClass_vtable.lookup_invokable = METHOD(VMClass, lookup_invokable);
    _VMClass_vtable.lookup_field_index = METHOD(VMClass, lookup_field_index);
    _VMClass_vtable.add_instance_invokable = 
        METHOD(VMClass, add_instance_invokable);
    _VMClass_vtable.add_instance_primitive =
        METHOD(VMClass, add_instance_primitive);
    _VMClass_vtable.get_instance_field_name =
        METHOD(VMClass, get_instance_field_name);
    _VMClass_vtable.get_number_of_instance_fields =
        METHOD(VMClass, get_number_of_instance_fields);
    _VMClass_vtable.has_primitives    = METHOD(VMClass, has_primitives);
    _VMClass_vtable.load_primitives   = METHOD(VMClass, load_primitives);

    VMClass_vtable_inited = true;
    
    return &_VMClass_vtable;
}
