/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */
 
#include "VMFrame.h"
#include "VMObject.h"
#include "VMMethod.h"
#include "VMClass.h"
#include "VMSymbol.h"
#include "VMInvokable.h"

#include <memory/gc.h>

#include <vm/Universe.h>

#include <interpreter/bytecodes.h>

#include <stddef.h>
#include <stdio.h>

//
//  Class Methods (Starting with VMFrame_) 
//
//

/**
 * Create a new VMFrame
 */
pVMFrame VMFrame_new(size_t length) {
    pVMFrame result = (pVMFrame)gc_allocate_object(
        sizeof(VMFrame) + sizeof(pVMObject) * length);
    if(result) {
        ((pVMFrame)DEREF(result))->_vtable = VMFrame_vtable();
        INIT(result, length);
    }
    return result;
}


pVMFrame VMFrame_new_with(pVMFrame previous_frame, pVMMethod method) {
    // Compute the maximum number of stack locations (including arguments,
    // locals and extra buffer to support doesNotUnderstand) and set the number
    // of indexable fields accordingly
    int length = SEND(method, get_number_of_arguments) +
                 SEND(method, get_number_of_locals) +
                 SEND(method, get_maximum_number_of_stack_elements) + 2;
    
    // Allocate a new frame and set its class to be the frame class
    pVMFrame result = VMFrame_new(length);
    SEND((pVMObject)result, set_class, frame_class);
    
    //SEND((pVMArray)result, set_number_of_indexable_fields, length);
    
    // Set the method of the frame and the previous frame
    SEND(result, set_method, method);
    if(previous_frame != NULL) 
        SEND(result, set_previous_frame, previous_frame);
    
    // Reset the stack pointer and the bytecode index
    SEND(result, reset_stack_pointer);
    SEND(result, set_bytecode_index, 0);
    
    // Return the freshly allocated frame
    return result;
}


/**
 * Initialize a VMFrame
 */
void _VMFrame_init(void* _self, ...) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    
    va_list args;
    va_start(args, _self);
    
    SUPER(VMArray, _self, init, va_arg(args, size_t));
    
    va_end(args);

    self->local_offset = VMInteger_new();
    self->bytecode_index = VMInteger_new();
    self->stack_pointer = VMInteger_new();
}


//
//  Instance Methods (Starting with _VMFrame_) 
//
//
pVMFrame _VMFrame_get_previous_frame(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return self->previous_frame;
}


void _VMFrame_set_previous_frame(void* _self, pVMFrame previous_frame) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    self->previous_frame = previous_frame;
}


void _VMFrame_clear_previous_frame(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    self->previous_frame = (pVMFrame)nil_object;
}


bool _VMFrame_has_previous_frame(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return (pVMObject)self->previous_frame != nil_object;
}


bool _VMFrame_is_bootstrap_frame(void* _self) {
    pVMFrame self = (pVMFrame)_self;
    return !SEND(self, has_previous_frame);
}


pVMFrame _VMFrame_get_context(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return self->context;
}


void _VMFrame_set_context(void* _self, pVMFrame context) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    self->context = context;
}


bool _VMFrame_has_context(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return (pVMObject)self->context != nil_object;
}


pVMFrame _VMFrame_get_context_level(void* _self, int level) {
    pVMFrame self = (pVMFrame)_self;
    pVMFrame current = self;
  
    // Iterate through the context chain until the given level is reached
    while(level > 0) {
        current = SEND(current, get_context);
        --level;
    }
    return current;
}


pVMFrame _VMFrame_get_outer_context(void* _self) {
    pVMFrame self = (pVMFrame)_self;
    pVMFrame current = self;
    while(SEND(current, has_context))
        current = SEND(current, get_context);
    return current;
}


pVMMethod _VMFrame_get_method(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return self->method;
}


void _VMFrame_set_method(void* _self, pVMMethod method) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    self->method = method;
}


pVMObject _VMFrame_pop(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    int sp = SEND(self->stack_pointer, get_embedded_integer);
    VMINTEGER_SET(self->stack_pointer, sp - 1);
    return SEND((pVMFrame)_self, get_indexable_field, sp);
}


void _VMFrame_push(void* _self, pVMObject value) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    size_t sp = SEND(self->stack_pointer, get_embedded_integer) + 1;
    VMINTEGER_SET(self->stack_pointer, sp);
    SEND((pVMFrame)_self, set_indexable_field, sp, value);
}


void _VMFrame_reset_stack_pointer(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    // arguments are stored in front of local variables
    pVMMethod meth = SEND((pVMFrame)_self, get_method);
    size_t lo = SEND(meth, get_number_of_arguments);
    VMINTEGER_SET(self->local_offset, lo);
  
    // Set the stack pointer to its initial value thereby clearing the stack
    size_t num_lo = SEND(meth, get_number_of_locals);
    VMINTEGER_SET(self->stack_pointer, lo + num_lo - 1);
}


int _VMFrame_get_bytecode_index(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    return SEND(self->bytecode_index, get_embedded_integer);
}


void _VMFrame_set_bytecode_index(void* _self, int index) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    VMINTEGER_SET(self->bytecode_index, index);
}


pVMObject _VMFrame_get_stack_element(void* _self, int index) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    size_t sp = SEND(self->stack_pointer, get_embedded_integer);
    return SEND((pVMFrame)_self, get_indexable_field, sp - index);
}


void _VMFrame_set_stack_element(void* _self, int index, pVMObject value) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    size_t sp = SEND(self->stack_pointer, get_embedded_integer);
    SEND((pVMFrame)_self, set_indexable_field, sp - index, value);    
}


pVMObject _VMFrame_get_local(void* _self, int index, int context_level) {
    pVMFrame self = (pVMFrame)_self;
    pVMFrame context = SEND(self, get_context_level, context_level);
    size_t lo = SEND(((pVMFrame)DEREF(context))->local_offset, get_embedded_integer);
    return SEND(context, get_indexable_field, lo + index);
}


void _VMFrame_set_local(void* _self, int index, int context_level,
    pVMObject value
) {
    pVMFrame self = (pVMFrame)_self;
    pVMFrame context = SEND(self, get_context_level, context_level);
    size_t lo = SEND(((pVMFrame)DEREF(context))->local_offset, get_embedded_integer);
    SEND(context, set_indexable_field, lo + index, value);
}


int _VMFrame_argument_stack_index(void* _self, int index) {
    pVMFrame self = (pVMFrame)_self;
    pVMMethod meth = SEND(self, get_method);
    return SEND(meth, get_number_of_arguments) - index - 1;
}


pVMObject _VMFrame_get_argument(void* _self, int index, int context_level) {
    pVMFrame self = (pVMFrame)_self;
    // get the context
    pVMFrame context = SEND(self, get_context_level, context_level);
  
    // get the argument with the given index
    return SEND(context, get_indexable_field, index);
}


void _VMFrame_set_argument(void* _self, int index, int context_level,
    pVMObject value
) {
    pVMFrame self = (pVMFrame)_self;
    // get the context
    pVMFrame context = SEND(self, get_context_level, context_level);
  
    // set the argument with the given index to the given value
    SEND(context, set_indexable_field, index, value);
}


void _VMFrame_copy_arguments_from(void* _self, pVMFrame frame) {
    pVMFrame self = (pVMFrame)_self;
    // copy arguments from frame:
    // - arguments are at the top of the stack of frame.
    // - copy them into the argument area of the current frame
    pVMMethod meth = SEND(self, get_method);
    int num_args = SEND(meth, get_number_of_arguments);
    for(int i=0; i < num_args; ++i) {
        pVMObject stackElem = SEND(frame, get_stack_element, num_args - 1 - i);
        SEND(self, set_indexable_field, i, stackElem);
    }
}


void _VMFrame_print_stack_trace(void* _self) {
    pVMFrame self = (pVMFrame)DEREF(_self);
    
    // retrieve frame and method data
    
    // method
    pVMMethod method = SEND((pVMFrame)_self, get_method);    
    pVMSymbol methodSym = TSEND(VMInvokable, method, get_signature);
    pString s_methodSym = SEND(methodSym, get_string);
    
    // holding class
    pVMClass holder = TSEND(VMInvokable, method, get_holder);
    pVMSymbol holderSym = SEND(holder, get_name);
    pString s_holderSym = SEND(holderSym, get_string);
    
    // sending bytecode
    size_t bc_idx = SEND(self->bytecode_index, get_embedded_integer);
    if(!SEND(self, is_bootstrap_frame)) 
        bc_idx -= 2; // length of SEND / SUPER_SEND
    bytecode_t bc = SEND(method, get_bytecode, bc_idx);

    // current selector, if any
    pString s_sel = NULL;  
    if(bc == BC_SEND || bc == BC_SUPER_SEND) {
        pVMSymbol sel = (pVMSymbol)SEND(method, get_constant, bc_idx);
        s_sel = SEND(sel, get_string);
    }
    
    debug_print("STACKTRACE:%20s>>%-20s%c@ %04d: %s%s\n",
                SEND(s_holderSym, chars),
                SEND(s_methodSym, chars),
                TSEND(VMInvokable, method, is_primitive) ? '*' : ' ',
                bc_idx,
                bytecodes_get_bytecode_name(bc),
                s_sel ? SEND(s_sel, chars) : "");
    
    // traverse contexts, if any
    if(SEND((pVMFrame)_self, has_previous_frame))
        SEND(self->previous_frame, print_stack_trace);    
}


size_t _VMFrame__get_offset(void* _self) {
    return (size_t)NUMBER_OF_OBJECT_FIELDS(VMFrame);
}


//
// The VTABLE-function
//
//
static VTABLE(VMFrame) _VMFrame_vtable;
bool VMFrame_vtable_inited = false;

VTABLE(VMFrame)* VMFrame_vtable(void) {
    if(VMFrame_vtable_inited)
        return &_VMFrame_vtable;
       
    *((VTABLE(VMArray)*)&_VMFrame_vtable) = *VMArray_vtable();
    _VMFrame_vtable.init               = METHOD(VMFrame, init);
    _VMFrame_vtable.get_previous_frame = METHOD(VMFrame, get_previous_frame);
    _VMFrame_vtable.set_previous_frame = METHOD(VMFrame, set_previous_frame);
    _VMFrame_vtable.clear_previous_frame = 
        METHOD(VMFrame, clear_previous_frame);
    _VMFrame_vtable.has_previous_frame =
        METHOD(VMFrame, has_previous_frame);
    _VMFrame_vtable.is_bootstrap_frame =
        METHOD(VMFrame, is_bootstrap_frame);
    _VMFrame_vtable.get_context = METHOD(VMFrame, get_context);
    _VMFrame_vtable.set_context = METHOD(VMFrame, set_context);
    _VMFrame_vtable.has_context = METHOD(VMFrame, has_context);
    _VMFrame_vtable.get_context_level = METHOD(VMFrame, get_context_level);
    _VMFrame_vtable.get_outer_context = METHOD(VMFrame, get_outer_context);
    _VMFrame_vtable.get_method = METHOD(VMFrame, get_method);
    _VMFrame_vtable.set_method = METHOD(VMFrame, set_method);
    _VMFrame_vtable.pop = METHOD(VMFrame, pop);
    _VMFrame_vtable.push = METHOD(VMFrame, push);
    _VMFrame_vtable.reset_stack_pointer = METHOD(VMFrame, reset_stack_pointer);
    _VMFrame_vtable.get_bytecode_index = METHOD(VMFrame, get_bytecode_index);
    _VMFrame_vtable.set_bytecode_index = METHOD(VMFrame, set_bytecode_index);
    _VMFrame_vtable.get_stack_element = METHOD(VMFrame, get_stack_element);
    _VMFrame_vtable.set_stack_element = METHOD(VMFrame, set_stack_element);
    _VMFrame_vtable.get_local = METHOD(VMFrame, get_local);
    _VMFrame_vtable.set_local = METHOD(VMFrame, set_local);
    _VMFrame_vtable.get_argument = METHOD(VMFrame, get_argument);
    _VMFrame_vtable.set_argument = METHOD(VMFrame, set_argument);
    _VMFrame_vtable.print_stack_trace = METHOD(VMFrame, print_stack_trace);
    _VMFrame_vtable.argument_stack_index = 
        METHOD(VMFrame, argument_stack_index);
    _VMFrame_vtable.copy_arguments_from = METHOD(VMFrame, copy_arguments_from);
    _VMFrame_vtable._get_offset = METHOD(VMFrame, _get_offset);

    VMFrame_vtable_inited = true;

    return &_VMFrame_vtable;
}




