
#include <stdio.h>
#include <unistd.h>

#include "Scheduler.h"

#include <module.VM>
#include <module.Bytecode>
#include <module.Interpreter>
#include <module.Memory>

#include <vmobjects/VMMethod.h>
#include <vmobjects/VMArray.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMSymbol.h>

// global objects
pVMClass scheduler_class;

//thread list
threadToFrameAssociation* frames = NULL;
//currently running thread
threadToFrameAssociation* currentThread = NULL;
//scheduler thread and frame for preemption
threadToFrameAssociation* schedulerThread = NULL;

threadToFrameAssociation* Scheduler_get_thread_frame_list() {
	return frames;
}

threadToFrameAssociation* Scheduler_get_scheduler_thread() {
	return schedulerThread;
}

bool Scheduler_firstRun = true;

//the dedicated scheduler frame
pVMFrame schedulerFrame = NULL;


pVMFrame Scheduler_get_scheduler_frame() {
	return schedulerFrame;
}

threadToFrameAssociation* Scheduler_get_current_thread() {
	return currentThread;
}
 
//holds whether the system is currently in an uninterubtable sections
bool uninterruptableSection = false;
 
//counts the entries and leavings if uninteruptable sections
int uninterruptable_section_count = 0;
 
//counts the local and non-local returns
size_t scheduler_return_count = 0;

//define the bytecodecount necessary to reinclude the scheduler
#define MAX_BYTECOUNT 1200


void Scheduler_enter_uninterruptable_section() {
    uninterruptable_section_count++;
	uninterruptableSection = true;   
}


void Scheduler_leave_uninterruptable_section() {
	uninterruptable_section_count--;
	//Set the boolean variable to false only if all previously
	//entered sections have been left
	if (uninterruptable_section_count == 0) {
		uninterruptableSection = false;  
	}
}

void Scheduler_set_scheduler_frame(pVMFrame frame) {
	schedulerFrame = frame;
}


void Scheduler_insert_scheduler() {
	if(!uninterruptableSection) {
		if(scheduler_return_count > MAX_BYTECOUNT && schedulerFrame != NULL) {
			//This should be unnecessary, but is left for safety reasons
			pVMMethod m = SEND(Interpreter_get_frame(), get_method);
			if(TSEND(VMInvokable, m, get_holder) != scheduler_class) {
				Scheduler_run_scheduler_thread();
				scheduler_return_count = 0;
			}
		}
	}
}


void Scheduler_run_thread(pVMObject thread, pVMFrame frame) {
	//save the scheduler thread only at the first run
	if (Scheduler_firstRun) {
		Scheduler_firstRun = false;
		schedulerThread = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		schedulerThread->thread = thread;
		schedulerThread->frame = frame;
		schedulerThread->next = NULL;
		Scheduler_set_scheduler_frame(frame);
	}
		 
	if (currentThread != NULL) {
		currentThread->frame = frame;
	}
		
	threadToFrameAssociation* frameAss = frames;
	while (frameAss->thread != thread) {
		if (frameAss->next == NULL) {
			return;
		}
		frameAss = frameAss->next;
	}
	
	currentThread = frameAss;
	Interpreter_set_frame(frameAss->frame);
}

void  Scheduler_register_thread_frame(pVMObject arg) {
	pVMClass cl = SEND(arg, get_class);

	//create surrounding frame - bootstrap frame
	pVMMethod bootstrap_method = Interpreter_create_bootstrap_method(cl);

	pVMFrame bootstrap_frame = VMFrame_new_with(NULL, (pVMMethod)bootstrap_method);
    SEND(bootstrap_frame, push, arg);

	//and set the run method as the method to invoke
	pVMInvokable runMethod = (pVMInvokable)SEND(cl, lookup_invokable, Universe_symbol_for_chars("run"));
	pVMFrame newFrame = VMFrame_new_with(bootstrap_frame, (pVMMethod)runMethod);
	SEND(newFrame, set_argument, 0, 0, arg);

	//create first linked list entry if not created yet
	if (frames == NULL) {
		frames = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		frames->thread = arg;
		frames->frame = newFrame;
		frames->next = NULL;
	}
	//or append the new association to the list
	else {
		threadToFrameAssociation* frame = frames->next;
		threadToFrameAssociation* prevFrame = frames;
		while (frame != NULL) {
			prevFrame = frame;
			frame = frame->next;
		}
		frame = (threadToFrameAssociation*) internal_allocate(sizeof(threadToFrameAssociation));
		frame->thread = arg;
		frame->frame = newFrame;
		frame->next = NULL;
		prevFrame->next = frame;
	}
}

void Scheduler_unregister_thread_frame(pVMObject thread) {
	threadToFrameAssociation* threadFrame = frames;
	threadToFrameAssociation* prevFrame = NULL;
	while (threadFrame != NULL && threadFrame->thread != thread) {
		prevFrame = threadFrame;
		threadFrame = threadFrame->next;
	}
	if (threadFrame != NULL) {
		if (prevFrame != NULL)
			prevFrame->next = threadFrame->next;
		threadFrame->frame = NULL;
		threadFrame->thread = NULL;
		threadFrame->next = NULL;
		//internal_free(threadFrame);
	}
}

void Scheduler_run_scheduler_thread() {
	if (currentThread != NULL) {
		currentThread->frame = Interpreter_get_frame();
	}

	currentThread = schedulerThread;
	Interpreter_set_frame(schedulerThread->frame);
}

void Scheduler_sleep(long duration) {
	sleep(duration);
}
