/*
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */

#include "hash.h"
#include "Hashmap.h"

#include "debug.h"

#include <module.Memory>
#include <module.VM>


//
// HashmapElem: class for representing hash map elements
//
bool _HashmapElem_key_equal_to(void* _self, void* other) {
    pHashmapElem self = (pHashmapElem)_self;
    return self->key == other;
}


int32_t _HashmapElem_key_hash(void* _self) {
    pHashmapElem self = (pHashmapElem)_self;
    return SEND(((pOOObject)self->key), hash_code);
}


pHashmapElem HashmapElem_new(void* k, void* v) {
    pHashmapElem result = (pHashmapElem)internal_allocate(sizeof(HashmapElem));
    if(result) {
        result->_vtable = HashmapElem_vtable();
        INIT2(result, (pOOObject)k, (pOOObject)v);
    }
    return result;
}


void _HashmapElem_init(void* _self, void* param1, void* param2) {
    pHashmapElem self = (pHashmapElem)_self;
    SUPER(OOObject, self, init, NULL, NULL);

    self->key = param1;
    self->value = param2;
}


void _HashmapElem_free(void* _self) {
    SUPER(OOObject, _self, free);
}

//
// Hashmap class
//
// The hash map implemented here uses a very simple collision handling scheme:
// in case a collision occurs, the next entry is tried, then the next, and so
// forth. This obviously is not the best implementation in terms of performance,
// but it keeps the implementation very simple.
//


//
//  Class Methods (Starting with Hashmap_) 
//


/**
 * Allocate a new Hashmap
 */
pHashmap Hashmap_new(void) {
    pHashmap result = (pHashmap)internal_allocate(sizeof(Hashmap));
    if(result) {
        result->_vtable = Hashmap_vtable();
        size_t size = HASHMAP_DEFAULT_SIZE;
        INIT1(result, &size);
    }    
    return result;    
}


/**
 * Set up the Hashmap object
 */
void _Hashmap_init(void* _self, void* param1, void* param2) {
    pHashmap self = (pHashmap)_self;
    SUPER(OOObject, self, init, NULL, NULL);

    self->size = *(size_t*)param1;
    self->elems = 
        (pHashmapElem*)internal_allocate(self->size * sizeof(pHashmapElem));
    SEND(self, clear);
}


//
//  Instance Methods (Starting with _Hashmap_) 
//


void _Hashmap_free(void* _self) {
    internal_free(((pHashmap)_self)->elems);
    SUPER(OOObject, _self, free);
}


/**
 * Get an element from a hash map
 */
void* _Hashmap_get(void* _self, void* key) {
    pHashmap self = (pHashmap)_self;
    return primitiveGet(self,
        key, hashf(self, SEND(((pOOObject)key), hash_code)));
}


/**
 *  Put an element into a hash map
 */
void _Hashmap_put(void* _self, void* key, void* value) {
    pHashmap self = (pHashmap)_self;
    pHashmapElem elem = HashmapElem_new(key, value);
    if(!primitivePut(self, elem, SEND(elem, key_hash)))
        // something is terribly wrong now, bail out
        Universe_error_exit("Hash map overflow after rehashing.\n");
}


void _Hashmap_rehash(void* _self, size_t newSize) {
    pHashmap self = (pHashmap)_self;
    size_t oldSize = self->size;
    pHashmapElem* oldElems = self->elems;
    self->size = newSize;
    self->elems = (pHashmapElem*)internal_allocate(
        self->size * sizeof(pHashmapElem));
    SEND(self, clear);
    
    for(size_t i = 0; i < oldSize; i++)
        if(oldElems[i]) {
            if(!attemptPut(self, oldElems[i],
                hashf(self, SEND(oldElems[i], key_hash))))
                // not good, can't help
                Universe_error_exit("Hashmap overflow while rehashing.\n");
            else
                oldElems[i] = NULL;
        }
            
    internal_free(oldElems);
}


/**
 *  Check existence of a given key in a hash map
 */
bool _Hashmap_contains_key(void* _self, void* key) {
    pHashmap self = (pHashmap)_self;
    return SEND(self, get, key) ? true : false;
}


/**
 *  Clear a hash map
 */
void _Hashmap_clear(void* _self) {
    pHashmap self = (pHashmap)_self;
    for(size_t i = 0; i < self->size; i++)
        self->elems[i] = NULL;
}


//
// vtable initialisation
//


static VTABLE(HashmapElem) _HashmapElem_vtable;
bool HashmapElem_vtable_inited = false;


VTABLE(HashmapElem)* HashmapElem_vtable(void) {
    if(!HashmapElem_vtable_inited) {
        *((VTABLE(OOObject)*)&_HashmapElem_vtable) = *OOObject_vtable();
        _HashmapElem_vtable.free         = METHOD(HashmapElem, free);
        _HashmapElem_vtable.init         = METHOD(HashmapElem, init);
        _HashmapElem_vtable.key_equal_to = METHOD(HashmapElem, key_equal_to);
        _HashmapElem_vtable.key_hash     = METHOD(HashmapElem, key_hash);            
        HashmapElem_vtable_inited = true;
    }
    return &_HashmapElem_vtable;
}


static VTABLE(Hashmap) _Hashmap_vtable;
bool Hashmap_vtable_inited = false;


VTABLE(Hashmap)* Hashmap_vtable(void) {
    if(!Hashmap_vtable_inited) {
        *((VTABLE(OOObject)*)&_Hashmap_vtable) = *OOObject_vtable();
        _Hashmap_vtable.free         = METHOD(Hashmap, free);
        _Hashmap_vtable.init         = METHOD(Hashmap, init);
        _Hashmap_vtable.get          = METHOD(Hashmap, get);
        _Hashmap_vtable.put          = METHOD(Hashmap, put);
        _Hashmap_vtable.clear        = METHOD(Hashmap, clear);
        _Hashmap_vtable.contains_key = METHOD(Hashmap, contains_key);
        _Hashmap_vtable.rehash       = METHOD(Hashmap, rehash);
        Hashmap_vtable_inited = true;
    }
    return &_Hashmap_vtable;
}
