/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
 
#include <module.Interpreter>
#include <module.Memory>
#include <module.VM>

#include "Shell.h"


#include <misc/Hashmap.h>
#include <misc/List.h>

#include <vmobjects/VMObject.h>
#include <vmobjects/VMFrame.h>
#include <vmobjects/VMMethod.h>
#include <vmobjects/VMArray.h>
#include <vmobjects/VMInvokable.h>
#include <vmobjects/VMBlock.h>
#include <vmobjects/VMBigInteger.h>
#include <vmobjects/VMInteger.h>
#include <vmobjects/VMDouble.h>
#include <vmobjects/VMString.h>
#include <vmobjects/VMClass.h>
#include <vmobjects/Symboltable.h>

#include <compiler/SourcecodeCompiler.h>

#include <misc/debug.h>

//
// values for debug output verbosity levels (controllable via the command line)
//


short dump_bytecodes;


// private helper functions
static int  setup_class_path(pString cp);
static int  add_class_path(const pString restrict cp);
static int  get_path_class_ext(pString **tokens, pString arg);

//file-local variables
static pString* class_path = NULL;
static int cp_count = 0;
static pHashmap globals_dictionary = NULL;

///////////////////////////////////
#pragma mark private hepler functions

static int get_path_class_ext(pString **tokens, pString arg) {   
#define EXT_TOKENS 2    
     
    // result String array
    *tokens = (pString *)internal_allocate(EXT_TOKENS * sizeof(pString));
    
    // index last path/file separator
    int fp_index = SEND(arg, lastIndexOfChar, *file_separator);
    // index of filename/suffix separator
    int ssep_index = SEND(arg, lastIndexOfChar, '.');
        
    if(fp_index < 0) { //no new path
        internal_free(*tokens);
        return ERR_FAIL;    
    } else //copy path token
        (*tokens)[0] = SEND(arg, substring, 0, fp_index - 1);
    
    // copy filename to 2nd slot
    // strip suffix if present
    ssep_index = ( (ssep_index >= 0)  && (ssep_index > fp_index)) ?
                 (ssep_index - 1):
                 SEND(arg, length);
    // copy substring
    (*tokens)[1] = SEND(arg, substring, fp_index + 1, ssep_index);
    
    return ERR_SUCCESS;
}


/*
 *  setup 'path_separator'-separated Classpathes
 *  from cp
 */
static int setup_class_path(const pString restrict cp) {
    /*
     * affected globals:
     * affected file globals: class_path, cp_count
     */
    
    size_t count;
    if ((class_path = SEND(cp, tokenize, &count, path_separator))) {
        cp_count = count;
        return ERR_SUCCESS;
    } else
        return ERR_NOMEM;
}


/**
 * add cp to the class_path vector
 */
static int add_class_path(const pString restrict cp) {
    /*
     * affected    globals:
     * affected file globals: class_path, cp_count
     */
    
    // resize class_path-vector
    class_path = (pString*)realloc(class_path, (cp_count + 1) * sizeof(pString));
    if(!class_path) 
        return ERR_NOMEM;
    // copy it and increment
    class_path[cp_count++] = cp;
    
    return ERR_SUCCESS;    
    
}


static void print_usage_options() {
    fprintf(stderr, "    -cp <directories separated by %s>\n", path_separator);
    fprintf(stderr, "        set search path for application classes\n");
    fprintf(stderr, "    -d  enable disassembling (twice for tracing)\n");
    fprintf(stderr, "    -h  show this help\n");
}


static void Universe_print_usage_and_exit(char* executable) {
    // print the usage
    fprintf(stderr, "Usage: %s [-options] [args...]\n\n", executable);
    fprintf(stderr, "where options include:\n");

	print_usage_options();

    // exit
    Universe_exit(ERR_SUCCESS);
}

////////////////

#pragma mark Extern Callable Functions

void Universe_exit(int err) {
    Universe_destruct();
    exit(err);
}


void Universe_error_exit(const char* restrict err) {
    debug_error("Runtime error: %s\n", err);
    Universe_exit(ERR_FAIL);
}

bool handle_complex_argument(char** argv, size_t* i, int argc) {
	if (strcmp(argv[*i], "-cp") == 0) { // Classpath
     	if ((argc == *i + 1) // no CP given
        	|| class_path) { // CP already set up.
                Universe_print_usage_and_exit(argv[0]);
        }

        // setup & skip class path
        setup_class_path(String_new(argv[++(*i)]));
	} else if ((strcmp(argv[*i], "-h") == 0) ||
        	(strcmp(argv[*i], "--help") == 0)) {
    	Universe_print_usage_and_exit(argv[0]);
	} else {
		return false;
	}
	return true;
}

bool handle_simple_argument(char* arg) {
	if (strcmp(arg, "-d") == 0) { // Dump  bytecode
        dump_bytecodes++;
    } else {
		return false;
	}
	return true;
}


pString* Universe_handle_arguments(int* vm_argc, int argc, char** argv) {
    //VM Arguments to return
    pString* vm_args = (pString*)internal_allocate(sizeof(pString*));
    *vm_argc = 0;
    
    dump_bytecodes = 0;
    
    /* iterate agruments (argv[>0])
     *
     */
    for (size_t i = 1; i < argc ; i++) {
        if (!handle_complex_argument(argv, &i, argc) && !handle_simple_argument(argv[i])) {
            /* copy remaining args for vm */
            
            // out of memory for the vm_args pointer
            if(!vm_args) Universe_exit(ERR_NOMEM);
                        
            // temp vector for path
            pString* ext_path_tokens = NULL;
            pString tmp_string = String_new(argv[i]);
            if(get_path_class_ext(&ext_path_tokens, tmp_string) ==
               ERR_SUCCESS
            ) {
                // got additional path, so add it
                add_class_path(ext_path_tokens[0]);
                                
                // copy filename to args
                /// TODO!!: resizing!!
                vm_args[(*vm_argc)++] = ext_path_tokens[1];

                // guaranteed not to be used/referenced any more
                memory_free(ext_path_tokens);
                
            } else {
                // no path, so copy arg only    
                // copy filename to args
                vm_args[(*vm_argc)++] = String_new_from(tmp_string);
            }
            // will not be referenced, because copied
            SEND(tmp_string, free);
        }
    }

    // Add local dir to class_path
    add_class_path(String_new("."));
    
    return vm_args;
}


void Universe_initialize(int argc, pString* argv) {
    /*
     * affected globals:
     * affected file globals: globals_dictionary
     */

    // setup the Hashmap for all globals
    globals_dictionary = Hashmap_new();
    // init the Symboltable
    Symbol_table_init();
    ///////////////////////////////////
    
    VMObjects_initialize();    
    
    // load methods and fields into the system classes
    Universe_load_system_class(object_class);
    Universe_load_system_class(class_class);
    Universe_load_system_class(metaclass_class);
    Universe_load_system_class(nil_class);
    Universe_load_system_class(array_class);
    Universe_load_system_class(method_class);
    Universe_load_system_class(symbol_class);
    Universe_load_system_class(integer_class);
    Universe_load_system_class(biginteger_class);
    Universe_load_system_class(frame_class);
    Universe_load_system_class(primitive_class);
    Universe_load_system_class(string_class);
    Universe_load_system_class(double_class);

    // load the generic block class
    block_class = Universe_load_class(Universe_symbol_for_chars("Block"));
    
    // setup the true and false objects
    true_object = VMObject_new_instance(Universe_load_class(Universe_symbol_for_chars("True")));
    false_object = VMObject_new_instance(Universe_load_class(Universe_symbol_for_chars("False")));
    
    // load the system class and create an instance of it
    system_class = Universe_load_class(Universe_symbol_for_chars("System"));
    pVMObject system_object = VMObject_new_instance(system_class);
    
    // put special objects and classes into the dictionary of globals
    Universe_set_global(Universe_symbol_for_chars("nil"), nil_object);
    Universe_set_global(Universe_symbol_for_chars("true"), true_object);
    Universe_set_global(Universe_symbol_for_chars("false"), false_object);
    Universe_set_global(Universe_symbol_for_chars("system"), system_object);
    Universe_set_global(Universe_symbol_for_chars("System"), (pVMObject)system_class);
    Universe_set_global(Universe_symbol_for_chars("Block"), (pVMObject)block_class);

    
    // create a fake bootstrap method to simplify later frame traversal
    pVMMethod bootstrap_method = Interpreter_create_bootstrap_method(system_class);
     
    // start the shell if no filename is given
    if(argc == 0) {
      Shell_set_bootstrap_method(bootstrap_method);
      Shell_start();
      return;
    }

    /* only trace bootstrap if the number of cmd-line "-d"s is > 2 */
    int trace = 2 - dump_bytecodes;
    if(!(trace > 0)) dump_bytecodes = 1;
    
    // convert the arguments into an array
    pVMArray arguments_array = VMArray_new_from_strings(argc, argv);
        
    // create a fake bootstrap frame with the system object on the stack
    pVMFrame bootstrap_frame = Interpreter_push_new_frame(bootstrap_method);
    SEND(bootstrap_frame, push, system_object);
    SEND(bootstrap_frame, push, (pVMObject)arguments_array);
            
    // lookup the initialize invokable on the system class
    pVMObject initialize =
        (pVMObject)SEND(system_class, lookup_invokable,
                        Universe_symbol_for_chars("initialize:"));
        
    // invoke the initialize invokable
    TSEND(VMInvokable, initialize, invoke, bootstrap_frame);

    // reset "-d" indicator
    if(!(trace>0)) dump_bytecodes = 2 - trace;
    
    // start the interpreter
    Interpreter_start();
}


/**
 * Universe Destructor
 * Free "member" variables
 */    
void Universe_destruct(void) {
    /*
     * affected    globals:
     * affected file globals: class_path, globals_dictionary
     */

    Symbol_table_destruct();
    if(globals_dictionary)
        SEND(globals_dictionary, free);
    FREE_ARRAY(class_path, cp_count);
}


void Universe_assert(bool value) {
    if(!value)
      // for now we just print something whenever an assertion fails
      debug_error("Assertion failed!\n");
}


pVMSymbol Universe_symbol_for(pString string) {
    // Lookup the symbol in the symbol table
    pVMSymbol result = Symbol_table_lookup(SEND(string, chars));
    
    // return found or newly created symbol
    return result ?
           result : // found
           VMSymbol_new_from_string(string); //new
}


pVMSymbol Universe_symbol_for_chars(const char* restrict string) {
    // Lookup the symbol in the symbol table
    pVMSymbol result = Symbol_table_lookup(string);
    
    // return found or newly created symbol
    return result ? result : VMSymbol_new_from_string(String_new(string)); //new
}


pHashmap Universe_get_globals_dictionary() {
    return globals_dictionary;
}


pVMObject Universe_get_global(pVMSymbol name) {
    // Return the global with the given name if it's in the dictionary of
    // globals
    if(Universe_has_global(name))
        return (pVMObject)SEND(globals_dictionary, get, name);
    
    // Global not found
    return NULL;
}


void Universe_set_global(pVMSymbol name, pVMObject value) {
    // Insert the given value into the dictionary of globals
    SEND(globals_dictionary, put, name, value);
}


bool Universe_has_global(pVMSymbol name) {
    // Returns if the Universe has a value for the global of the given name
    return SEND(globals_dictionary, contains_key, name);
}


pVMClass Universe_get_block_class(void) {
    return block_class;
}


pVMClass Universe_get_block_class_with_args(size_t number_of_arguments) {
    // Compute the name of the block class with the given number of arguments
    char block_name[7];
    Universe_assert(number_of_arguments <10); // buffer overflow otherwise
    sprintf(block_name, "Block%d", number_of_arguments);
    pVMSymbol name = Universe_symbol_for_chars(block_name);
    
    // Lookup the specific block class in the dictionary of globals and return
    // it
    if(Universe_has_global(name)) 
        return (pVMClass)Universe_get_global(name);

    // Get the block class for blocks with the given number of arguments
    pVMClass result = Universe_load_class_basic(name, NULL);
    
    // Add the appropriate value primitive to the block class
    SEND(result, add_instance_primitive,
         VMBlock_get_evaluation_primitive(number_of_arguments));
    
    // Insert the block class into the dictionary of globals
    Universe_set_global(name, (pVMObject)result);
    
    // Return the loaded block class
    return result;
}


pVMClass Universe_load_class(pVMSymbol name) {
    // Check if the requested class is already in the dictionary of globals
    if(Universe_has_global(name)) 
        return (pVMClass)Universe_get_global(name);
    
    // Load the class
    pVMClass result = Universe_load_class_basic(name, NULL);
        
    // check class loading.
    if(!result) {
        pString name_s = SEND(name, get_string);
        debug_error("can't load class:\t'%s'\n", SEND(name_s, chars));
        Universe_exit(ERR_FAIL);
    }

    // Load primitives (if necessary) and return the resulting class
    if(SEND(result, has_primitives) || SEND(result->clazz, has_primitives)) 
        SEND(result, load_primitives, class_path, cp_count);
    return result;
}


void Universe_load_system_class(pVMClass system_class) {
    // Load the system class
    pVMClass result =
        Universe_load_class_basic(SEND(system_class, get_name), system_class);
    
    // check class loading.
    if(!result) {
        pVMSymbol cname = SEND(system_class, get_name);
        debug_error("can't load system class:\t%s", SEND(cname, get_string));
        Universe_exit(ERR_FAIL);
    }

    // Load primitives (if necessary) 
    if(SEND(result, has_primitives) || SEND(result->clazz, has_primitives)) 
        SEND(result, load_primitives, class_path, cp_count);
}


pVMClass Universe_load_class_basic(pVMSymbol name, pVMClass system_class) {
    pString s_name = SEND(name, get_string);
    debug_log("%s\n", SEND(s_name, chars));

    pVMClass result;
    // Try loading the class from all different paths
    for(int i = 0; i < cp_count; i++) {
        // Load the class from a file and return the loaded class
        result = SourcecodeCompiler_compile_class(class_path[i], 
                                                  SEND(name, get_string),
                                                  system_class);
        if(result) {
            if(dump_bytecodes) {
                Disassembler_dump(SEND(result, get_class));
                Disassembler_dump(result);
            }
            // found class in class_path, so return it
            return result; 
        }
    }
    // The class could not be found.
    return NULL;
}


pVMClass Universe_load_shell_class(pString stmt) {
    // Load the class from a stream and return the loaded class
    pVMClass result = SourcecodeCompiler_compile_class_string(stmt, NULL);
    if(dump_bytecodes)
        Disassembler_dump(result);
    return result;    
}
