/*
 *
Copyright (c) 2007 Michael Haupt, Tobias Pape
Software Architecture Group, Hasso Plattner Institute, Potsdam, Germany
http://www.hpi.uni-potsdam.de/swa/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
  */

#include "VMArray.h"
#include "VMString.h"

#include <module.Memory>
#include <module.VM>
#include <module.VMObjects>

#include <misc/List.h>


//
//  Class Methods (Starting with VMArray_) 
//


/** 
 *  Create a new VMArray
 *
 */
pVMArray VMArray_new(size_t size) {
    pVMArray result = (pVMArray)memory_allocate_object(
        sizeof(VMArray) + sizeof(pVMObject) * size);
    if(result) {
        result->_vtable = VMArray_vtable();
        INIT1(result, &size);
    }
    return result;
}


pVMArray VMArray_new_from_list(pList list) {
    // Allocate a new array with the same length as the list
    pVMArray result = VMArray_new(SEND(list, size));
    
    if(result) {
        // Copy all elements from the list into the array
        for(int i = 0; i < SEND(list, size); i++) {
            pVMObject elem =  (pVMObject)SEND(list, get, i);
            SEND(result, set_indexable_field, i, elem);
        }
    }    
    // Return the allocated and initialized array
    return result;
}


pVMArray VMArray_new_from_strings(int count, pString* strings) {
    // allocate a new array with the same length as the string array
    pVMArray result = VMArray_new(count);
    // copy all elements from the string array into the array

    for(int i = 0; i < count; i++) {
        SEND(result, set_indexable_field, i,
            (pVMObject)VMString_new_with(strings[i]));
    }

    // return the allocated and initialized array
    return result;
}

/** 
 *
 * Initialize a VMArray
 *
 */
void _VMArray_init(void* _self, void* param1, void* param2) {
    pVMArray self = (pVMArray)_self;
    
    size_t fields = SEND(self, _get_offset) + *(size_t*)param1;
    SUPER(VMObject, _self, init, &fields, NULL);
    
    SEND(self, set_class, array_class);
}


//
// Special instance Method: must be overridden by every subclass
//

/**
 * Return the offset of the indexable Fields from "normal" fields
 */
size_t _VMArray__get_offset(void* _self) {
    return NUMBER_OF_OBJECT_FIELDS(VMArray);
}


//
//  Instance Methods (Starting with _VMArray_) 
//


#define CHECK_INDEX \
    if(index >= SEND(self, get_number_of_indexable_fields) || index < 0) { \
        char* s = (char*)internal_allocate(128); \
        if(index >= SEND(self, get_number_of_indexable_fields)) \
            sprintf(s, "index too large in array access: %d", index); \
        else \
            sprintf(s, "index too small in array access: %d", index); \
        Universe_error_exit(s); \
    }


pVMObject _VMArray_get_indexable_field(void* _self, intptr_t index) {
    pVMArray self = (pVMArray)_self;
    //CHECK_INDEX;
    if(index >= SEND(self, get_number_of_indexable_fields) || index < 0) {
        char* s = (char*)internal_allocate(128);
        if(index >= SEND(self, get_number_of_indexable_fields))
            sprintf(s, "index too large in array access: %d", index);
        else
            sprintf(s, "index too small in array access: %d", index);
        Universe_error_exit(s);
    }
    // get the indexable field with the given index
    return self->fields[index + SEND(self, _get_offset)];
}


void _VMArray_set_indexable_field_without_bounds_check(pVMArray self, int index, pVMObject value) {
    // set the indexable field with the given index to the given value
    self->fields[index] = value;
}

void _VMArray_set_indexable_field(void* _self, intptr_t index, pVMObject value) {
    pVMArray self = (pVMArray)_self;
    //CHECK_INDEX;
    if(index >= SEND(self, get_number_of_indexable_fields) || index < 0) {
        char* s = (char*)internal_allocate(128);
        if(index >= SEND(self, get_number_of_indexable_fields))
            sprintf(s, "index too large in array access: %d", index);
        else
            sprintf(s, "index too small in array access: %d", index);
        Universe_error_exit(s);
    }

    _VMArray_set_indexable_field_without_bounds_check(self, index + SEND(self, _get_offset), value);
}


int _VMArray_get_number_of_indexable_fields(void* _self) {
    pVMArray self = (pVMArray)_self;
    // retrieve the number of indexable fields in this array
    return self->num_of_fields - SEND(self, _get_offset);
}


pVMArray _VMArray_copy_and_extend_with(void* _self, pVMObject value) {
    pVMArray self = (pVMArray)_self;
    size_t fields = SEND(self, get_number_of_indexable_fields);
    // allocate a new array which has one indexable field more than this array
    pVMArray result = VMArray_new(fields + 1);
    // copy the indexable fields from this array to the new array
    SEND(self, copy_indexable_fields_to, result);
    // insert the given object as the last indexable field in the new array
    SEND(result, set_indexable_field, fields, value);
    // return the new array
    return result;
}


void _VMArray_copy_indexable_fields_to(void* _self, pVMArray destination) {
    pVMArray self = (pVMArray)_self;
    size_t fields = SEND(self, get_number_of_indexable_fields);
    // copy all indexable fields from this array to the destination array
    for(int i = 0; i < fields; i++) {
        pVMObject object = SEND(self, get_indexable_field, i);
        SEND(destination, set_indexable_field, i, object);
    }
}
